"""Base model."""

from ansys.api.edb.v1.edb_messages_pb2 import EDBObjMessage

from ansys.edb.core.utility.io_manager import get_buffer, get_cache, get_io_manager


class ObjBase:
    """Provides the base object that all gRPC-related models extend from."""

    def __init__(self, msg):
        """Initialize the base object.

        Parameters
        ----------
        msg : EDBObjMessage
        """
        self.msg = msg

    @property
    def is_null(self):
        """:obj:`bool`: Flag indicating if the object exists in the database.

        This property is read-only.
        """
        if (buffer := get_buffer()) is not None:
            buffer.flush()
        return self.id == 0

    @property
    def id(self):
        """:obj:`int`: Unique ID of the EDB object.

        A ``0`` indicates an invalid object.

        This property is read-only.
        """
        return self._id

    @property
    def msg(self):
        """:obj:`EDBObjMessage`: Protobuf message that represents the object's ID.

        This property can only be set to :obj:`None`.
        """
        msg = EDBObjMessage(id=self.id)
        io_mgr = get_io_manager()
        if io_mgr.is_enabled:
            if self._is_future:
                msg.is_future = True
            io_mgr.active_request_edb_obj_msg_mgr.add_active_request_edb_obj_msg(msg)
        return msg

    @msg.setter
    def msg(self, msg):
        if msg is None:
            self._id = 0
            return
        self._id = msg.id
        self._is_future = msg.is_future
        if self._is_future:
            if (buffer := get_buffer()) is not None:
                buffer.add_future_ref(self)
        else:
            if (cache := get_cache()) is not None:
                cache.add_from_cache_msg(msg)


class TypeField(object):
    """Provides a descriptor for a type field that can be overridden by subclasses.

    You can have an optional ``@property def _type(self)`` in the same class as a fallback method to
    fetch the type from the server in case the static type is unknown.

    Examples
    --------
    class PrimitiveType(Enum):
        CIRCLE = 0

    class Primitive(object):
        type = TypeField(None)
        @property
        def _type(self):
            return PrimitiveType(self.__stub.GetPrimitiveType())

    class Circle(Primitive):
        type = TypeField(PrimitiveType.CIRCLE)

    Primitive(...).type # => fetch type from server.
    Primitive.type      # => None.
    Circle(...).type    # => CIRCLE.
    Circle.type         # => CIRCLE.
    """

    def __init__(self, tp):
        """Initialize a type field with static concrete type."""
        self._type = tp

    def __get__(self, instance, owner):
        """Get the static type if possible. Otherwise, fetch and return the instance type."""
        if self._type is not None:
            # return static type
            return self._type
        if instance is not None and hasattr(instance, "_type"):
            # fetch type
            return instance._type  # noqa
        else:
            # no type info available
            return None



"""Pin pair model."""

import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.hierarchy.model import Model
from ansys.edb.core.inner import messages, parser
from ansys.edb.core.session import PinPairModelServiceStub, StubAccessor, StubType


class PinPairModel(Model):
    """Represents a pin pair model object."""

    __stub: PinPairModelServiceStub = StubAccessor(StubType.pin_pair_model)

    @classmethod
    def create(cls):
        """Create a pin pair model.

        Returns
        -------
        PinPairModel
            Pin pair model created.
        """
        return cls(cls.__stub.Create(empty_pb2.Empty()))

    def rlc(self, pin_pair):
        """Get the RLC value for a pin pair.

        Parameters
        ----------
        pin_pair : tuple[str, str]

        Returns
        -------
        :class:`.Rlc`
        """
        res = self.__stub.GetRlc(messages.string_pair_property_message(self, pin_pair))
        if res.found:
            return parser.to_rlc(res.rlc)
        else:
            return None

    def set_rlc(self, pin_pair, rlc):
        """Set the RLC value for a pin pair.

        Parameters
        ----------
        pin_pair : tuple[str, str]
        rlc : :class:`.Rlc`
        """
        self.__stub.SetRlc(messages.pin_pair_model_rlc_message(self, pin_pair, rlc))

    def delete_rlc(self, pin_pair):
        """Delete the RLC value for a pin pair.

        Parameters
        ----------
        pin_pair : tuple[str, str]
        """
        self.__stub.DeleteRlc(messages.string_pair_property_message(self, pin_pair))

    def pin_pairs(self):
        """Get all pin pairs.

        Returns
        -------
        list[tuple[str, str]]
        """
        res = self.__stub.GetPinPairs(messages.edb_obj_message(self))
        return [(pair.first, pair.second) for pair in res.pairs]



"""This package contains common type definitions used throughout the EDB codebase."""

from typing import Iterable, Tuple, Union

import ansys.edb.core.geometry.point_data as point_data
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.net.net import Net
from ansys.edb.core.utility.value import Value

ValueLike = Union[int, float, complex, str, Value]
PointLike = Union[point_data.PointData, Iterable[ValueLike]]
Point3DLike = Tuple[ValueLike, ValueLike, ValueLike]
Triangle3DLike = Tuple[Point3DLike, Point3DLike, Point3DLike]
LayerLike = Union[Layer, str]
NetLike = Union[Net, str]



"""Text."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.typing import LayerLike, ValueLike


from ansys.api.edb.v1 import text_pb2, text_pb2_grpc

from ansys.edb.core.inner import messages
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class Text(Primitive):
    """Represents a text object."""

    __stub: text_pb2_grpc.TextServiceStub = StubAccessor(StubType.text)

    @classmethod
    def create(
        cls, layout: Layout, layer: LayerLike, center_x: ValueLike, center_y: ValueLike, text: str
    ) -> Text:
        """Create a text object.

        Parameters
        ----------
        layout : .Layout
            Layout to create the text object in.
        layer : :term:`LayerLike`
            Layer to place the text object on.
        center_x : :term:`ValueLike`
            X value of the center point.
        center_y : :term:`ValueLike`
            Y value of the center point.
        text : str
            Text string.

        Returns
        -------
        Text
            Text object created.
        """
        return Text(
            cls.__stub.Create(
                text_pb2.TextCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    text=text,
                )
            )
        )

    def get_text_data(self) -> tuple[Value, Value, str]:
        """Get the data for the text object.

        Returns
        -------
        tuple of (.Value, .Value, str)
            Returns a tuple in this format:

            **(center_x, center_y, text)**

            **center_x** : X value of center point.

            **center_y** : Y value of center point.

            **radius** : Text object's String value.
        """
        text_data_msg = self.__stub.GetTextData(self.msg)
        return (
            Value(text_data_msg.center_x),
            Value(text_data_msg.center_y),
            text_data_msg.text,
        )

    def set_text_data(self, center_x: ValueLike, center_y: ValueLike, text: str):
        """Set the data for the text object.

        Parameters
        ----------
        center_x : :term:`ValueLike`
            X value of the center point.
        center_y : :term:`ValueLike`
            Y value of the center point.
        text : str
            String value for the text object.
        """
        self.__stub.SetTextData(
            text_pb2.SetTextDataMessage(
                target=self.msg,
                data=text_pb2.TextDataMessage(
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    text=text,
                ),
            )
        )



"""Log module.

## Objective
This module intends to create a general framework for logging.
This module is built upon the ``logging`` library and it does NOT intend to
replace it but rather provide a way to interact between ``logging`` and ``EDB Client``.

Usage
-----

Global logger
-------------

There is a global logger named ``EDB_LOGGER``.
If you want to use this global logger, you must call it at the top of your module:

.. code::
    from ansys.edb.core.utility import EDB_LOG

The default logging level of ``LOG`` is ``ERROR``.
To change this and output lower-level messages, you can use this code:

.. code::
   EDB_LOG.setLevel(logging.DEBUG)

By default, this logger does not log to a file.
If you want to log to a file, can add a file handler using this code:

.. code::
   import os
   file_path = os.path.join(os.getcwd(), 'edb_client.log')
   EDB_LOG.log_to_file(file_path)

The preceding code also sets the logger to be redirected to the file.

To log using this logger, call the desired method as a normal logger:

.. code::
    import logging
    from ansys.edb.core.utility import edb_logging
    LOGGER = EDBLogger(level=logging.DEBUG, to_file=False, to_stdout=True)
    LOGGER.debug('This is LOG debug message.')

    | DEBUG    | This is LOG debug message.
"""

from datetime import datetime
import logging
import sys

# Default configuration
LOG_LEVEL = logging.DEBUG
FILE_NAME = "edb_client.log"

# Formatting

STDOUT_MSG_FORMAT = "%(levelname)s - %(message)s"

FILE_MSG_FORMAT = STDOUT_MSG_FORMAT

DEFAULT_STDOUT_HEADER = """
"""
DEFAULT_FILE_HEADER = DEFAULT_STDOUT_HEADER

NEW_SESSION_HEADER = f"""
===============================================================================
       NEW SESSION - {datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}
==============================================================================="""


class EDBLogger:
    """Provides the EDB logger used for each EDC client session.

    This class allows you to add a handler to a file or to the standard output.
    """

    file_handler = None
    std_out_handler = None

    def __init__(self, level=logging.DEBUG, to_file=False, to_stdout=True, filename=FILE_NAME):
        """Customize logger class for the EDB cclient.

        Parameters
        ----------
        level : int, default: DEBUG
            Level of logging as defined in the ``logging`` package.
        to_file : bool, default: False
            Whether to record the logs in a file.
        to_stdout : bool, default: True
            Whether to output the logs to the standard output, which is the
            command line.
        filename : str, default: FILE_NAME
            Name of the output file. The default is ``FILE_NAME, in which case
            ``edb_client.log`` is used.
        """
        self.logger = logging.getLogger("edb_client")  # Creating default main logger.
        self.logger.setLevel(level)
        self.logger.propagate = True

        if to_file or filename != FILE_NAME:
            # Record to file
            self.log_to_file(filename=filename, level=level)

        if to_stdout:
            self.log_to_stdout(level=level)

        add_handling_uncaught_exceptions(self.logger)  # Use logger to record unhandled exceptions

    def __getattr__(self, item):
        """Delegate method calls to the logger."""
        return getattr(self.logger, item)

    def stop_logging_to_stdout(self):
        """Stop logging to the standard output."""
        if self.std_out_handler is not None:
            self.logger.removeHandler(self.std_out_handler)
            self.std_out_handler = None

    def stop_logging_to_file(self):
        """Stop logging to the file."""
        if self.file_handler is not None:
            self.logger.removeHandler(self.file_handler)
            self.file_handler = None

    def log_to_file(self, filename=FILE_NAME, level=LOG_LEVEL):
        """Add file handler to the logger.

        Parameters
        ----------
        filename : str, default: FILE_NAME
            Name of the file where logs are recorded.
        level : str, default: default: LOG_LEVEL
            Level of logging, such as ``DEBUG``.
        """
        addfile_handler(self, filename=filename, level=level, write_headers=True)

    def log_to_stdout(self, level=LOG_LEVEL):
        """Add the standard output handler to the logger.

        Parameters
        ----------
        level : str, default: LOG_LEVEL
            Level of logging record.
        """
        add_stdout_handler(self, level=level)


# Auxiliary functions


def add_handling_uncaught_exceptions(logger):
    """Redirect the output of an exception to the logger."""

    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        logger.critical("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = handle_exception


def addfile_handler(edb_logger, filename=FILE_NAME, level=LOG_LEVEL, write_headers=False):
    """
    Add a file handler to the input.

    Parameters
    ----------
    edb_logger : EDBLogger
        EDB logger to add the file handler to.
    filename : str, default: FILE_NAME.
        Name of the output file.
    level : str, default: LOG_LEVEL
        Level of log recording.
    write_headers : bool, default: False
        Whether to write the headers to the file.

    Returns
    -------
    logger
        EDBLogger object.
    """
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(logging.Formatter(FILE_MSG_FORMAT))

    if isinstance(edb_logger, EDBLogger):
        edb_logger.file_handler = file_handler
        edb_logger.logger.addHandler(file_handler)

    if write_headers:
        file_handler.stream.write(NEW_SESSION_HEADER)
        file_handler.stream.write(DEFAULT_FILE_HEADER)

    return edb_logger


def add_stdout_handler(edb_logger, level=LOG_LEVEL, write_headers=False):
    """Add a file handler to the logger.

    Parameters
    ----------
    edb_logger : EDBLogger
        EDBLogger to add the file handler to.
    level : str, default: LOG_LEVEL
        Level of log recording. The default is ``LOG_LEVEL``, in which
        case ``logging.DEBUG`` is used.
    write_headers : bool, default: False
        Whether to write the headers to the file.

    Returns
    -------
    logger
        EDBLogger object.
    """
    std_out_handler = logging.StreamHandler()
    std_out_handler.setLevel(level)
    std_out_handler.setFormatter(logging.Formatter(STDOUT_MSG_FORMAT))

    if isinstance(edb_logger, EDBLogger):
        edb_logger.std_out_handler = std_out_handler
        edb_logger.logger.addHandler(std_out_handler)

    if write_headers:
        std_out_handler.stream.write(DEFAULT_STDOUT_HEADER)

    return edb_logger



"""RPC Info Utils."""

from ansys.edb.core.inner.rpc_info import rpc_information


def get_rpc_info(service_name, rpc_name):
    """Get the rpc info for the rpc corresponding to the provided service and rpc names."""
    buffered_service = rpc_information.get(service_name)
    return buffered_service.get(rpc_name) if buffered_service is not None else None


def can_cache(service_name, rpc_name):
    """Check if the response of the rpc corresponding to the provided service and rpc names can be cached."""
    if (rpc_info := get_rpc_info(service_name, rpc_name)) is not None:
        return rpc_info.can_cache
    return False


def can_buffer(service_name, rpc_name):
    """Check if the request of the rpc corresponding to the provided service and rpc names can be buffered."""
    if (rpc_info := get_rpc_info(service_name, rpc_name)) is not None:
        return rpc_info.can_buffer
    return False


def is_read(service_name, rpc_name):
    """Check if the rpc corresponding to the provided service and rpc names is a read operation."""
    if (rpc_info := get_rpc_info(service_name, rpc_name)) is not None:
        return rpc_info.is_read
    return False


def is_write(service_name, rpc_name):
    """Check if the rpc corresponding to the provided service and rpc names is a write operation."""
    if (rpc_info := get_rpc_info(service_name, rpc_name)) is not None:
        return rpc_info.is_write
    return False



"""This module parses messages back to client data types."""

import functools

from ansys.edb.core.simulation_setup.adaptive_solutions import (
    AdaptiveFrequency,
    BroadbandAdaptiveSolution,
    MatrixConvergenceData,
    MultiFrequencyAdaptiveSolution,
    SingleFrequencyAdaptiveSolution,
)
from ansys.edb.core.simulation_setup.mesh_operation import (
    LengthMeshOperation,
    SkinDepthMeshOperation,
)


def to_point_data(fn):
    """Decorate a function that returns a message to return it as a ``PointData`` object."""
    return _wraps(fn, msg_to_point_data)


def to_point_data_list(fn):
    """Decorate a function that returns a message to return it as a list of ``PointData`` objects."""
    return _wraps(fn, _to_point_data_list)


def to_point3d_data(fn):
    """Decorate a function that returns a message to return it as a ``Point3DData`` object."""
    return _wraps(fn, _to_point3d_data)


def to_point_data_pair(fn):
    """Decorate a function that returns a message to return it as a ``[PointData, PointData]`` tuple."""
    return _wraps(fn, _to_point_data_pair)


def to_3_point3d_data(fn):
    """Decorate a function that returns a message to return it as a list of ``Point3DData`` objects."""
    return _wraps(fn, _to_3_point3d_data)


def to_polygon_data(fn):
    """Decorate a function that returns a message to return it as a ``PolygonData`` object."""
    return _wraps(fn, msg_to_polygon_data)


def to_polygon_data_list(fn):
    """Decorate a function that returns a message to return it as a list of ``PolygonData`` objects."""
    return _wraps(fn, _to_polygon_data_list)


def to_rlc(fn):
    """Decorate a function that returns a message to return it as an RLC."""
    return _wraps(fn, _to_rlc)


def to_box(fn):
    """Decorate a function that returns a message to return it as ``(lower_left, upper_right)``."""
    return _wraps(fn, _to_box)


def to_circle(fn):
    """Decorate a function that returns a message to return it as ``(center, radius)``."""
    return _wraps(fn, _to_circle)


def to_base_adaptive_frequency_solution(fn):
    """Decorate a function that returns a message to return it as a ``[`str`, `str`]`` tuple."""
    return _wraps(fn, _to_base_adaptive_frequency_solution)


def to_single_frequency_adaptive_solution(fn):
    """Decorate a function that returns a message to return it as a ``SingleFrequencyAdaptiveSolution`` object."""
    return _wraps(fn, _to_single_frequency_adaptive_solution)


def to_multi_adaptive_freq(fn):
    """Decorate a function that returns a message to return it as an ``AdaptiveFrequency`` object."""
    return _wraps(fn, _to_multi_adaptive_freq)


def to_multi_frequency_adaptive_solution(fn):
    """Decorate a function that returns a message to return it as a ``MultiFrequencyAdaptiveSolution`` object."""
    return _wraps(fn, _to_multi_frequency_adaptive_solution)


def to_broadband_adaptive_solution(fn):
    """Decorate a function that returns a message to return it as a ``BroadbandAdaptiveSolution`` object."""
    return _wraps(fn, _to_broadband_adaptive_solution)


def to_mesh_op(fn):
    """Decorate a function that returns a message to return it as a ``MeshOperation`` object."""
    return _wraps(fn, _to_mesh_op)


def to_string_dict(fn):
    """Decorate a function that returns a message to return it as a ``dict[str, str]``."""
    return _wraps(fn, _to_string_dict)


def _wraps(fn, wrapper_fn):
    if callable(fn):

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            return wrapper_fn(fn(*args, **kwargs))

        return wrapper
    else:
        return wrapper_fn(fn)


def msg_to_point_data(message):
    """Convert a ``PointMessage`` object to a ``PointData`` object.

    Parameters
    ----------
    message : ansys.api.edb.v1.point_data_pb2.PointMessage

    Returns
    -------
    :class:`.PointData`
    """
    from ansys.edb.core.geometry.point_data import PointData
    from ansys.edb.core.utility.value import Value

    return PointData([Value(message.x), Value(message.y)])


def _to_point_data_pair(message):
    """Convert a ``PointPairMessage`` object to a ``[PointData, PointData]`` tuple.

    Parameters
    ----------
    message : ansys.api.edb.v1.point_data_pb2.PointPairMessage

    Returns
    -------
    tuple[:class:`.PointData`, :class:`.PointData`]
    """
    return msg_to_point_data(message.point_0), msg_to_point_data(message.point_1)


def _to_point_data_list(message):
    """Convert a message to list of ``PointData`` objects.

    Parameters
    ----------
    message : list[ansys.api.edb.v1.point_data_pb2.PointMessage]

    Returns
    -------
    list[:class:`.PointData`]
    """
    return [msg_to_point_data(m) for m in message]


def _to_3_point3d_data(message):
    """Convert a ``Point3DMessage`` object to a ``PointData`` object.

    Parameters
    ----------
    message : list[ansys.api.edb.v1.point_data_pb2.CPos3DTripleMessage]

    Returns
    -------
    :class:`.Point3DData`
    """
    return [_to_point3d_data(message.x), _to_point3d_data(message.y), _to_point3d_data(message.z)]


def _to_point3d_data(message):
    """Convert a ``Point3DMessage`` object to a ``PointData`` object.

    Parameters
    ----------
    message : ansys.api.edb.v1.point_data_pb2.Point3DMessage

    Returns
    -------
    :class:`.Point3DData`
    """
    from ansys.edb.core.geometry.point3d_data import Point3DData
    from ansys.edb.core.utility.value import Value

    return Point3DData(Value(message.x), Value(message.y), Value(message.z))


def msg_to_polygon_data(message):
    """Convert an arbitrary message to a ``PolygonData`` object if possible.

    Parameters
    ----------
    message : ansys.api.edb.v1.point_data_pb2.BoxMessage or ansys.api.edb.v1.polygon_data_pb2.PolygonDataMessage

    Returns
    -------
    :class:`.PolygonData`
    """
    from ansys.api.edb.v1.point_data_pb2 import BoxMessage

    from ansys.edb.core.geometry.polygon_data import PolygonData

    if isinstance(message, BoxMessage):
        b = _to_box(message)
        return PolygonData(lower_left=b[0], upper_right=b[1])
    else:
        return PolygonData(
            points=_to_point_data_list(message.points),
            holes=_to_polygon_data_list(message.holes),
            sense=message.sense,
            closed=message.closed,
        )


def _to_polygon_data_list(message):
    """Convert arbitrary messages to list of ``PolygonData`` objects if possible.

    Returns
    -------
    list[:class:`.PolygonData`]
    """
    if hasattr(message, "polygons"):
        return [msg_to_polygon_data(m) for m in message.polygons]
    elif hasattr(message, "points"):
        return [msg_to_polygon_data(m) for m in message.points]
    else:
        return [msg_to_polygon_data(m) for m in message]


def _to_box(message):
    """Convert a message to box.

    Parameters
    ----------
    message : ansys.api.edb.v1.point_data_pb2.BoxMessage

    Returns
    -------
    tuple[:class:`.PointData`, :class:`.PointData`]
    """
    if hasattr(message, "lower_left") and hasattr(message, "upper_right"):
        return msg_to_point_data(message.lower_left), msg_to_point_data(message.upper_right)


def _to_circle(message):
    """Convert a message to a circle containing a center point and radius.

    Parameters
    ----------
    message : ansys.api.edb.v1.point_data_pb2.CircleMessage

    Returns
    -------
    tuple[:class:`.PointData`, :class:`.Value`]
    """
    from ansys.edb.core.utility.value import Value

    if hasattr(message, "center") and hasattr(message, "radius"):
        return msg_to_point_data(message.center), Value(message.radius)


def _to_rlc(message):
    """Convert a message to an RLC containing values related to resistance inductance capacitance.

    Parameters
    ----------
    message : ansys.api.edb.v1.rlc_pb2.RlcMessage

    Returns
    -------
    :class:`.Rlc`
    """
    from ansys.edb.core.utility.rlc import Rlc
    from ansys.edb.core.utility.value import Value

    return Rlc(
        Value(message.r),
        message.r_enabled.value,
        Value(message.l),
        message.l_enabled.value,
        Value(message.c),
        message.c_enabled.value,
        message.is_parallel.value,
    )


def _to_mx_convergence_data(message):
    """Convert a message to matrix convergence data.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_settings_pb2.MatrixConvergenceDataMessage

    Returns
    -------
    MatrixConvergenceData
    """
    mx_entry_msgs = message.entries
    if len(mx_entry_msgs) == 0:
        return

    mx_conv_data = MatrixConvergenceData()
    if (
        message.all_are_constant
        or message.all_diag_are_constant
        or message.all_off_diag_are_constant
    ):
        port_name_set = set()
        port_name_list = []

        def add_port_name_entry(port_name):
            if port_name not in port_name_set:
                port_name_set.add(mx_entry_msg.port_1)
                port_name_list.append(mx_entry_msg.port_1)

        for mx_entry_msg in mx_entry_msgs:
            add_port_name_entry(mx_entry_msg.port_1)
            add_port_name_entry(mx_entry_msg.port_2)

        if message.all_are_constant:
            first_msg = mx_entry_msgs[0]
            mx_conv_data.set_all_constant(
                first_msg.mag_limit, first_msg.phase_limit, port_name_list
            )

        if message.all_diag_are_constant:
            for mx_entry_msg in mx_entry_msgs:
                if mx_entry_msg.port_1 == mx_entry_msg.port_2:
                    mx_conv_data.set_all_diag_constant(
                        mx_entry_msg.mag_limit, mx_entry_msg.phase_limit, port_name_list, True
                    )
                    break

        if message.all_off_diag_are_constant:
            for mx_entry_msg in mx_entry_msgs:
                if mx_entry_msg.port_1 != mx_entry_msg.port_2:
                    mx_conv_data.set_all_off_diag_constant(
                        mx_entry_msg.mag_limit, mx_entry_msg.phase_limit, port_name_list, False
                    )
                    break
    else:
        for mx_entry_msg in mx_entry_msgs:
            mx_conv_data.add_entry(
                mx_entry_msg.port_1,
                mx_entry_msg.port_2,
                mx_entry_msg.mag_limit,
                mx_entry_msg.phase_limit,
            )

    return mx_conv_data


def _to_base_adaptive_frequency_solution(message):
    """Convert a message to adaptive frequency solution data.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_settings_pb2.AdaptiveFrequencyDataMessage

    Returns
    -------
    tuple[`str`, `str`]
    """
    return message.adaptive_frequency, message.max_delta


def _to_single_frequency_adaptive_solution(message):
    """Convert a message to single frequency adaptive solution data.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_settings_pb2.SingleFrequencyAdaptiveSolutionMessage

    Returns
    -------
    SingleFrequencyAdaptiveSolution
    """
    adaptive_freq_data = _to_base_adaptive_frequency_solution(message.adaptive_frequency)
    return SingleFrequencyAdaptiveSolution(
        adaptive_freq_data[0],
        adaptive_freq_data[1],
        message.max_passes,
        _to_mx_convergence_data(message.matrix_conv_data),
        message.use_matrix_conv_data,
    )


def _to_multi_adaptive_freq(message):
    """Convert a message to a multi-frequency adaptive solution adaptive frequency entry.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_settings_pb2.SingleFrequencyAdaptiveSolutionMessage

    Returns
    -------
    AdaptiveFrequency
    """
    adaptive_freq_data = _to_base_adaptive_frequency_solution(message.adaptive_frequency)
    return AdaptiveFrequency(
        adaptive_freq_data[0],
        adaptive_freq_data[1],
        {key: value for (key, value) in message.output_variables},
    )


def _to_multi_frequency_adaptive_solution(message):
    """Convert a message to a multi-frequency adaptive solution data.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_settings_pb2.MultiFrequencyAdaptiveSolutionMessage

    Returns
    -------
    MultiFrequencyAdaptiveSolution
    """
    return MultiFrequencyAdaptiveSolution(
        message.max_passes,
        [_to_multi_adaptive_freq(freq_msg) for freq_msg in message.adaptive_frequencies],
    )


def _to_broadband_adaptive_solution(message):
    """Convert a message to broadband adaptive solution data.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_settings_pb2.BroadbandFrequencyAdaptiveSolutionMessage

    Returns
    -------
    BroadbandAdaptiveSolution
    """
    return BroadbandAdaptiveSolution(
        message.low_frequency, message.high_frequency, message.max_passes, message.max_delta
    )


def _length_mesh_op(message):
    """Convert a message to a length mesh operation.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_setup_pb2.LengthMeshOperationMessage

    Returns
    -------
    LengthMeshOperation
    """
    return LengthMeshOperation(
        max_length=message.max_length,
        restrict_max_length=message.restrict_length,
        max_elements=message.max_elements,
        restrict_max_elements=message.restrict_max_elements,
    )


def _to_skin_depth_mesh_op(message):
    """Convert a message to a skin depth mesh operation.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_setup_pb2.SkinDepthMeshOperationMessage

    Returns
    -------
    SkinDepthMeshOperation
    """
    return SkinDepthMeshOperation(
        skin_depth=message.skin_depth,
        surface_triangle_length=message.surface_triangle_length,
        num_layers=message.num_layers,
        max_elements=message.max_elements,
        restrict_max_elements=message.restrict_max_elements,
    )


def _to_mesh_op(message):
    """Convert a message to a mesh operation.

    Parameters
    ----------
    message : ansys.api.edb.v1.hfss_simulation_setup_pb2.MeshOperationMessage

    Returns
    -------
    MeshOperation
    """
    mesh_op = None
    if message.HasField("skin_depth_mesh_op"):
        mesh_op = _to_skin_depth_mesh_op(message.skin_depth_mesh_op)
    elif message.HasField("length_mesh_op"):
        mesh_op = _length_mesh_op(message.length_mesh_op)
    mesh_op.name = message.name
    mesh_op.enabled = message.enabled
    mesh_op.mesh_region = message.mesh_region
    mesh_op.refine_inside = message.refine_inside
    mesh_op.solve_inside = message.solve_inside
    for nli in message.net_layer_info:
        mesh_op.net_layer_info.append((nli.net, nli.layer, nli.is_sheet))
    return mesh_op


def _to_string_dict(message):
    """Convert a message to a dictionary of strings."""
    return {key: value for (key, value) in message.string_map.items()}



"""Voltage regulator."""
import ansys.api.edb.v1.voltage_regulator_pb2 as vr_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import conn_obj, messages
from ansys.edb.core.primitive.padstack_instance import PadstackInstance
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class PowerModule:
    """Represents a power module.

    Attributes
    ----------
    comp_group_name : str
        Component group name.
    pos_output_terminal : str
        Name of the positive output terminal.
    neg_output_terminal : str
        Name of the negative output terminal
    relative_strength : :class:`.Value`
        Relative strength as a percentage value.
    active : bool
        Whether the power module is active.
    """

    def __init__(
        self,
        comp_group_name,
        pos_output_terminal="",
        neg_output_terminal="",
        relative_strength=Value(100),
        active=True,
    ):
        """Construct a power module."""
        self._comp_group_name = comp_group_name
        self._pos_output_terminal = pos_output_terminal
        self._neg_output_terminal = neg_output_terminal
        self._relative_strength = Value(relative_strength)
        self._active = active
        self._needs_sync = True

    @property
    def comp_group_name(self):
        """:obj:`str`: Component group name of the power module."""
        return self._comp_group_name

    @comp_group_name.setter
    def comp_group_name(self, comp_group_name):
        self._comp_group_name = comp_group_name

    @property
    def pos_output_terminal(self):
        """:obj:`str`: Positive output terminal name for the power module."""
        return self._pos_output_terminal

    @pos_output_terminal.setter
    def pos_output_terminal(self, pos_output_terminal):
        self._pos_output_terminal = pos_output_terminal

    @property
    def neg_output_terminal(self):
        """:obj:`str`: Negative output terminal name for the power module."""
        return self._neg_output_terminal

    @neg_output_terminal.setter
    def neg_output_terminal(self, neg_output_terminal):
        self._neg_output_terminal = neg_output_terminal

    @property
    def relative_strength(self):
        """:class:`.Value`: Relative strength for the power module as a percentage.

        This property can be set with :term:`ValueLike`.
        """
        return self._relative_strength

    @relative_strength.setter
    def relative_strength(self, relative_strength):
        self._relative_strength = relative_strength

    @property
    def active(self):
        """:obj:`bool`: Flag indicating if the power module is active."""
        return self._active

    @active.setter
    def active(self, active):
        self._active = active

    @property
    def needs_sync(self):
        """:obj:`bool`: Flag indicating if the power module needs to be synchronized.

        This property is read-only.
        """
        return self._needs_sync


class VoltageRegulator(conn_obj.ConnObj):
    """Represents a voltage regulator."""

    __stub = StubAccessor(StubType.voltage_regulator)
    layout_obj_type = LayoutObjType.VOLTAGE_REGULATOR

    @classmethod
    def create(cls, layout, name, active, voltage, lrc, lrp):
        """
        Create a voltage regulator.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the voltage regulator in.
        name : str
            Name of the voltage regulator.
        active : bool
            Active state of the voltage regulator.
        voltage : :term:`ValueLike`
            Voltage of the voltage regulator.
        lrc : :term:`ValueLike`
            Load regulation current.
        lrp : :term:`ValueLike`
            Load regulation percentage.

        Returns
        -------
        VoltageRegulator
            Voltage regulator created.
        """
        return VoltageRegulator(
            cls.__stub.Create(
                vr_pb2.VoltageRegulatorMessage(
                    layout=layout.msg,
                    name=name,
                    active=active,
                    voltage=messages.value_message(voltage),
                    lrc=messages.value_message(lrc),
                    lrp=messages.value_message(lrp),
                )
            )
        )

    @property
    def name(self):
        """:obj:`str`: Name of the voltage regulator."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, newname):
        self.__stub.SetName(messages.edb_obj_name_message(self.msg, newname))

    @property
    def active(self):
        """:obj:`bool`: Flag indicating if the voltage regular is active."""
        return self.__stub.IsActive(self.msg).value

    @active.setter
    def active(self, newactive):
        self.__stub.SetIsActive(messages.bool_property_message(self, newactive))

    @property
    def voltage(self):
        """:class:`.Value`: Voltage of the voltage regulator.

        This property can be set with :term:`ValueLike`.
        """
        return Value(self.__stub.GetVoltage(self.msg))

    @voltage.setter
    def voltage(self, newvoltage):
        self.__stub.SetVoltage(
            messages.value_property_message(self.msg, messages.value_message(newvoltage))
        )

    @property
    def lrc(self):
        """:class:`.Value`: Load regulation current of the voltage regulator.

        This property can be set with :term:`ValueLike`.
        """
        return Value(self.__stub.GetLoadRegulationCurrent(self.msg))

    @lrc.setter
    def lrc(self, newlrc):
        self.__stub.SetLoadRegulationCurrent(
            messages.value_property_message(self.msg, messages.value_message(newlrc))
        )

    @property
    def lrp(self):
        """:class:`.Value`: Load regulation percent of the voltage regulator.

        This property can be set with :term:`ValueLike`.
        """
        return Value(self.__stub.GetLoadRegulationPercent(self.msg))

    @lrp.setter
    def lrp(self, newlrp):
        self.__stub.SetLoadRegulationPercent(
            messages.value_property_message(self.msg, messages.value_message(newlrp))
        )

    @property
    def pos_remote_sense_pin(self):
        """:class:`.PadstackInstance`: \
        Positive remote sense pin of the voltage regulator.

        .. seealso:: :obj:`neg_remote_sense_pin`
        """
        return PadstackInstance(self.__stub.GetPosRemoteSensePin(self.msg))

    @pos_remote_sense_pin.setter
    def pos_remote_sense_pin(self, newpin):
        self.__stub.SetPosRemoteSensePin(messages.edb_obj_collection_message([self, newpin]))

    @property
    def neg_remote_sense_pin(self):
        """:class:`.PadstackInstance`: \
        Negative remote sense pin of the voltage regulator.

        .. seealso:: :obj:`pos_remote_sense_pin`
        """
        return PadstackInstance(self.__stub.GetNegRemoteSensePin(self.msg))

    @neg_remote_sense_pin.setter
    def neg_remote_sense_pin(self, newpin):
        self.__stub.SetNegRemoteSensePin(messages.edb_obj_collection_message([self, newpin]))

    @property
    def num_power_modules(self):
        """:obj: `int`: Number of power modules.

        Read-Only
        """
        return self.__stub.GetNPowerModules(self.msg).value

    @property
    def num_active_power_modules(self):
        """:obj: `int`: Number of active power modules.

        This attribute is read-only
        """
        return self.__stub.GetNActivePowerModules(self.msg).value

    def get_power_module(self, comp_group_name):
        """Get the power module for a given component group name.

        Parameters
        ----------
        comp_group_name : str
            Component group name of the power module.

        Returns
        -------
        PowerModule
        """
        return VoltageRegulator._create_power_module(
            msg=self.__stub.GetPowerModule(messages.string_property_message(self, comp_group_name))
        )

    def get_all_power_modules(self):
        """Get all power modules in the voltage regulator.

        Returns
        -------
        list[PowerModule]
            List of all power modules.
        """
        all_pms = self.__stub.GetAllPowerModules(self.msg)

        return [VoltageRegulator._create_power_module(msg=msg) for msg in all_pms.data]

    def add_power_module(self, power_module):
        """Add a power module to the voltage regulator.

        Parameters
        ----------
        power_module : PowerModule
            Power module.
        """
        self.__stub.AddPowerModule(
            vr_pb2.PowerModulePropertyMessage(
                vrm=self.msg, module=messages.power_module_message(power_module)
            )
        )

    def remove_power_module(self, name):
        """Remove a power module from the voltage regulator.

        Parameters
        ----------
        name : str
            Component group name of the power module.
        """
        self.__stub.RemovePowerModule(messages.string_property_message(target=self, value=name))

    def add_power_modules(self, power_modules):
        """Add multiple power modules to the voltage regulator.

        Parameters
        ----------
        power_modules : list[PowerModule]
            List of power modules to add.
        """
        self.__stub.AddPowerModules([messages.power_module_message(pm) for pm in power_modules])

    def remove_power_modules(self, names):
        """Remove multiple power modules from the voltage regulator.

        Parameters
        ----------
        names : list[str]
            List of component group names of each power module to remove.

        """
        self.__stub.RemovePowerModules(messages.strings_property_message(target=self, value=names))

    def remove_all_power_modules(self):
        """Remove all power modules in the voltage regulator."""
        self.__stub.RemoveAllPowerModules(self.msg)

    @staticmethod
    def _create_power_module(msg):
        return PowerModule(
            comp_group_name=msg.comp_group_name,
            pos_output_terminal=msg.pos_output_terminal,
            neg_output_terminal=msg.neg_output_terminal,
            relative_strength=msg.relative_strength,
            active=msg.active,
        )



"""Port postprocessing properties."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike

from ansys.edb.core.utility.value import Value


class PortPostProcessingProp:
    """Represents the port postprocessing properties."""

    def __init__(
        self,
        voltage_magnitude: ValueLike = 0,
        voltage_phase: ValueLike = 0,
        deembed_length: ValueLike = 0,
        renormalization_impedance: ValueLike = 50,
        do_deembed: bool = False,
        do_deembed_gap_l: bool = False,
        do_renormalize: bool = False,
    ):
        """Initialize port postprocessing properties.

        Parameters
        ----------
        voltage_magnitude : :term:`ValueLike`, optional
            Excitation voltage magnitude. The default is 0.
        voltage_phase : :term:`ValueLike`, optional
            Excitation voltage phase.  The default is 0.
        deembed_length : :term:`ValueLike`, optional
            Dembeed distance.  The default is 0. This parameter is only
            applied if ``do_deembed=True``.
        renormalization_impedance : :term:`ValueLike`, optional
            Renormalization impedance. The default is 0. This parameter is only
            applied if ``do_renormalize=True``.
        do_deembed : bool, optional
            Whether to enable the port to be deembedded. The default is ``False``.
        do_deembed_gap_l : bool, optional
            Whether to enable port impedance renormalization. The default is ``False``.
        do_renormalize : bool, optional
            Whether to enable the gap port inductance to be deembedded.  The default
            is ``False``.
        """
        self.voltage_magnitude = voltage_magnitude
        self.voltage_phase = voltage_phase
        self.deembed_length = deembed_length
        self.renormalization_impedance = renormalization_impedance
        self.do_deembed = do_deembed
        """Enable port to be deembedded.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        self.do_deembed_gap_l = do_deembed_gap_l
        """Enable port impedance renormalization.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        self.do_renormalize = do_renormalize
        """Enable port impedance renormalization.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """

    @property
    def voltage_magnitude(self) -> Value:
        """:class:`.Value`: Excitation voltage magnitude.

        This property can be set to :term:`ValueLike`.
        """
        return self._voltage_magnitude

    @voltage_magnitude.setter
    def voltage_magnitude(self, value: ValueLike):
        self._voltage_magnitude = Value(value)

    @property
    def voltage_phase(self):
        """:class:`.Value`: Excitation voltage phase.

        This property can be set to :term:`ValueLike`.
        """
        return self._voltage_phase

    @voltage_phase.setter
    def voltage_phase(self, value: ValueLike):
        self._voltage_phase = Value(value)

    @property
    def deembed_length(self):
        """:class:`.Value`: Deembed length.

        This property can be set to :term:`ValueLike`.
        This property is only applied if ``do_deembed=True``.
        """
        return self._deembed_length

    @deembed_length.setter
    def deembed_length(self, value: ValueLike):
        self._deembed_length = Value(value)

    @property
    def renormalization_impedance(self):
        """:class:`.Value`: Renormalization impedance.

        This property can be set to :term:`ValueLike`.
        This property is only applied if ``do_renormalize=True``.
        """
        return self._renormalization_impedance

    @renormalization_impedance.setter
    def renormalization_impedance(self, value):
        self._renormalization_impedance = Value(value)



"""Layer."""

from enum import Enum

import ansys.api.edb.v1.layer_pb2 as layer_pb2
from ansys.api.edb.v1.layer_pb2_grpc import LayerServiceStub

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import ObjBase
from ansys.edb.core.inner.messages import (
    get_product_property_ids_message,
    get_product_property_message,
    set_product_property_message,
)
from ansys.edb.core.session import StubAccessor, StubType


# Message creation helper method
def _is_in_zone_message(lyr, zone):
    """Convert to an ``IsInZoneMessage`` object."""
    return layer_pb2.ZoneMessage(layer=lyr.msg, zone=zone)


class LayerType(Enum):
    """Provides an enum representing the types of layers."""

    SIGNAL_LAYER = layer_pb2.SIGNAL_LAYER
    DIELECTRIC_LAYER = layer_pb2.DIELECTRIC_LAYER
    CONDUCTING_LAYER = layer_pb2.CONDUCTING_LAYER
    AIRLINES_LAYER = layer_pb2.AIRLINES_LAYER
    ERRORS_LAYER = layer_pb2.ERRORS_LAYER
    SYMBOL_LAYER = layer_pb2.SYMBOL_LAYER
    MEASURE_LAYER = layer_pb2.MEASURE_LAYER
    ASSEMBLY_LAYER = layer_pb2.ASSEMBLY_LAYER
    SILKSCREEN_LAYER = layer_pb2.SILKSCREEN_LAYER
    SOLDER_MASK_LAYER = layer_pb2.SOLDER_MASK_LAYER
    SOLDER_PASTE_LAYER = layer_pb2.SOLDER_PASTE_LAYER
    GLUE_LAYER = layer_pb2.GLUE_LAYER
    WIREBOND_LAYER = layer_pb2.WIREBOND_LAYER
    USER_LAYER = layer_pb2.USER_LAYER
    SIWAVE_HFSS_SOLVER_REGIONS = layer_pb2.SIWAVE_HFSS_SOLVER_REGIONS
    POST_PROCESSING_LAYER = layer_pb2.POST_PROCESSING_LAYER
    OUTLINE_LAYER = layer_pb2.OUTLINE_LAYER
    LAYER_TYPES_COUNT = layer_pb2.LAYER_TYPES_COUNT
    UNDEFINED_LAYER_TYPE = layer_pb2.UNDEFINED_LAYER_TYPE


class TopBottomAssociation(Enum):
    """Provides an enum representing the top-bottom association of layers."""

    TOP_ASSOCIATED = layer_pb2.TOP_ASSOCIATED
    NO_TOP_BOTTOM_ASSOCIATED = layer_pb2.NO_TOP_BOTTOM_ASSOCIATED
    BOTTOM_ASSOCIATED = layer_pb2.BOTTOM_ASSOCIATED
    TOP_BOTTOM_ASSOCIATION_COUNT = layer_pb2.TOP_BOTTOM_ASSOCIATION_COUNT
    INVALID_TOP_BOTTOM_ASSOCIATION = layer_pb2.INVALID_TOP_BOTTOM_ASSOCIATION


class DrawOverride(Enum):
    """Provides an enum representing draw override options for layers."""

    NO_OVERRIDE = layer_pb2.NO_OVERRIDE
    FILL = layer_pb2.FILL
    WIREFRAME = layer_pb2.WIREFRAME


class LayerVisibility(Enum):
    """Provides an enum representing visibility options for layers."""

    PRIMITIVE_VISIBLE = layer_pb2.PRIMITIVE_VISIBLE
    PATH_VISIBLE = layer_pb2.PATH_VISIBLE
    PAD_VISIBLE = layer_pb2.PAD_VISIBLE
    HOLE_VISIBLE = layer_pb2.HOLE_VISIBLE
    COMPONENT_VISIBLE = layer_pb2.COMPONENT_VISIBLE
    ALL_VISIBLE = layer_pb2.ALL_VISIBLE


class Layer(ObjBase):
    """Provides the base class representing a layer."""

    layout_obj_type = LayoutObjType.LAYER
    __stub: LayerServiceStub = StubAccessor(StubType.layer)

    def cast(self):
        """Cast the layer object to the correct concrete type.

        Returns
        -------
        Layer
        """
        from ansys.edb.core.layer.stackup_layer import StackupLayer
        from ansys.edb.core.layer.via_layer import ViaLayer

        lyr = Layer(self.msg)
        if lyr.is_stackup_layer:
            if lyr.is_via_layer:
                return ViaLayer(self.msg)
            else:
                return StackupLayer(self.msg)
        else:
            return lyr

    @staticmethod
    def create(name, lyr_type):
        """Create a non-stackup layer.

        Parameters
        ----------
        name : string
            Layer name.
        lyr_type : :class:`LayerType`
            Layer type.

        Returns
        -------
        Layer
        """
        return Layer(
            Layer.__stub.Create(layer_pb2.LayerCreationMessage(name=name, type=lyr_type.value))
        )

    @property
    def type(self):
        """:class:`LayerType`: Type of the layer."""
        return LayerType(self.__stub.GetLayerType(self.msg).type)

    @type.setter
    def type(self, lyr_type):
        self.__stub.SetLayerType(layer_pb2.SetLayerTypeMessage(layer=self.msg, type=lyr_type.value))

    @property
    def is_stackup_layer(self):
        """:obj:`bool`: Flag indicating if the layer is a :class:`StackupLayer` instance.

        This property is read-only.
        """
        layer_type = self.type
        return (
            layer_type == LayerType.DIELECTRIC_LAYER
            or layer_type == LayerType.CONDUCTING_LAYER
            or layer_type == LayerType.SIGNAL_LAYER
        )

    @property
    def is_via_layer(self):
        """:obj:`bool`: Flag indicating if the layer is a :class:`ViaLayer` instance.

        This property is read-only.
        """
        return self.__stub.IsViaLayer(self.msg).value

    @property
    def name(self):
        """:obj:`str`: Name of the layer."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, name):
        self.__stub.SetName(layer_pb2.SetNameMessage(layer=self.msg, name=name))

    def clone(self, copy_id=True):
        """Create a clone of the layer.

        Parameters
        ----------
        copy_id : bool
            ID of the layer to clone.

        Returns
        -------
        Layer
            Layer cloned.
        """
        return Layer(self.__stub.Clone(layer_pb2.CloneMessage(layer=self.msg, copy_id=copy_id)))

    @property
    def layer_id(self):
        """:obj:`int`: Layer ID."""
        return self.__stub.GetLayerId(self.msg).value

    @property
    def top_bottom_association(self):
        """:class:`TopBottomAssociation`: Top-bottom association of the layer."""
        return TopBottomAssociation(
            self.__stub.GetTopBottomAssociation(self.msg).top_bottom_association
        )

    @top_bottom_association.setter
    def top_bottom_association(self, top_bottom_association):
        self.__stub.SetTopBottomAssociation(
            layer_pb2.SetTopBottomAssociationMessage(
                layer=self.msg, top_bottom_association=top_bottom_association.value
            )
        )

    @property
    def color(self):
        r""":obj:`tuple`\[:obj:`int`, :obj:`int`, :obj:`int`\]: Color of the layer in (R,G,B) format.

        Tuple contains the color values in (R,G,B) format.
        """
        color_int = self.__stub.GetColor(self.msg).value
        r = color_int & 0x000000FF
        g = (color_int & 0x0000FF00) >> 8
        b = (color_int & 0x00FF0000) >> 16
        return r, g, b

    @color.setter
    def color(self, rgb):
        r = rgb[0] & 0x000000FF
        g = (rgb[1] << 8) & 0x0000FF00
        b = (rgb[2] << 16) & 0x00FF0000
        self.__stub.SetColor(layer_pb2.SetColorMessage(layer=self.msg, color=b | g | r))

    @property
    def visibility_mask(self):
        """:obj:`int`: Visibility mask of the layer.

        The setter can take either an :obj:`int` object or a :class:`LayerVisibility` instance.
        """
        return self.__stub.GetVisibilityMask(self.msg).value

    @visibility_mask.setter
    def visibility_mask(self, visibility_mask):
        vis_mask_int = (
            visibility_mask.value
            if isinstance(visibility_mask, LayerVisibility)
            else visibility_mask
        )
        self.__stub.SetVisibilityMask(
            layer_pb2.SetVisibilityMaskMessage(layer=self.msg, visibility_mask=vis_mask_int)
        )

    @property
    def locked(self):
        """:obj:`bool`: Flag indicating if the layer is locked."""
        return self.__stub.GetLocked(self.msg).value

    @locked.setter
    def locked(self, locked):
        self.__stub.SetLocked(layer_pb2.SetLockedMessage(layer=self.msg, is_locked=locked))

    @property
    def transparency(self):
        """:obj:`int`: Transparency value of the layer.

        The transparency value is between 0 and 100, where 0 indicates a completely
        opaque layer and 100 indicates a completely transparent layer.
        """
        return self.__stub.GetTransparency(self.msg).value

    @transparency.setter
    def transparency(self, transparency):
        self.__stub.SetTransparency(
            layer_pb2.SetTransparencyMessage(layer=self.msg, transparency=transparency)
        )

    @property
    def draw_override(self):
        """:class:`DrawOverride`: Draw override of the layer."""
        return DrawOverride(self.__stub.GetDrawOverride(self.msg).draw_override)

    @draw_override.setter
    def draw_override(self, draw_override):
        self.__stub.SetDrawOverride(
            layer_pb2.SetDrawOverrideMessage(layer=self.msg, draw_override=draw_override.value)
        )

    def get_product_property(self, prod_id, attr_it):
        """Get the product property of the layer for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            Product ID.
        attr_it : int
            Attribute ID.

        Returns
        -------
        str
            Product property.
        """
        return self.__stub.GetProductProperty(
            get_product_property_message(self, prod_id, attr_it)
        ).value

    def set_product_property(self, prod_id, attr_it, prop_value):
        """Set the product property of the layer for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            Product ID.
        attr_it : int
            Attribute ID.
        prop_value : str
            New product property value.
        """
        self.__stub.SetProductProperty(
            set_product_property_message(self, prod_id, attr_it, prop_value)
        )

    def get_product_property_ids(self, prod_id):
        """Get a list of attribute IDs for a given product ID for the layer.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            Product ID.

        Returns
        -------
        list[int]
            List of attribute IDs.
        """
        attr_ids = self.__stub.GetProductPropertyIds(
            get_product_property_ids_message(self, prod_id)
        ).ids
        return [attr_id for attr_id in attr_ids]

    def is_in_zone(self, zone):
        """Determine if the layer exists in the given zone.

        Parameters
        ----------
        zone : int

        Returns
        -------
        bool
            ``True`` when the layer exists in the given zone, ``False`` otherwise.
        """
        return self.__stub.IsInZone(_is_in_zone_message(self, zone)).value

    def set_is_in_zone(self, zone, in_zone=True):
        """Set whether the layer exists in a given zone.

        Parameters
        ----------
        zone : int
           Zone.
        in_zone : bool, default: True
           Whether the layer exists in this zone.
        """
        return self.__stub.SetIsInZone(
            layer_pb2.SetIsInZoneMessage(zone_msg=_is_in_zone_message(self, zone), in_zone=in_zone)
        )

    @property
    def zones(self):
        r""":obj:`list`\[:obj:`int`\]: IDs of all zones containing the layer.

        This property is read-only.
        """
        return [zone for zone in self.__stub.GetZones(self.msg).zones]

    @property
    def zone(self):
        """:obj:`int`: Zone index associated with the owning layer collection.

        If the owner is invalid, the index is ``0``. If the owner is multizone,
        the index is ``-1``.

        This property is read-only.
        """
        return self.__stub.GetZone(self.msg).value



"""Layout instance context."""

from ansys.edb.core.inner import ObjBase, parser
from ansys.edb.core.inner.messages import bool_property_message
from ansys.edb.core.layout import layout
from ansys.edb.core.session import LayoutInstanceContextServiceStub, StubAccessor, StubType


class LayoutInstanceContext(ObjBase):
    """Represents the layout instance context object."""

    __stub: LayoutInstanceContextServiceStub = StubAccessor(StubType.layout_instance_context)

    @property
    def layout(self):
        """:class:`.Layout`: Layout of the context.

        This property is read-only.
        """
        return layout.Layout(self.__stub.GetLayout(self.msg))

    @parser.to_polygon_data
    def get_bbox(self, local):
        """Get the bounding box of the context.

        Parameters
        ----------
        local : bool
            Whether to return the bounding box in the local :class:`context <LayoutInstanceContext>`.
            If ``False``, the bounding-box in the global context is returned.

        Returns
        -------
        :class:`.PolygonData`
            Bounding box of the context.
        """
        return self.__stub.GetBBox(bool_property_message(self, local))

    @property
    def is_top_or_black_box(self):
        """:obj:`bool`: Flag indicating if this is a top-level or blackbox context.

        This property is read-only.
        """
        return self.__stub.IsTopOrBlackBox(self.msg).value

    @property
    def top_or_black_box(self):
        """:class:`LayoutInstanceContext`: Top-level or blackbox :class:`context <LayoutInstanceContext>` instance.

        This property is read-only.
        """
        return LayoutInstanceContext(self.__stub.GetTopOrBlackBox(self.msg))

    @property
    def placement_elevation(self):
        """:obj:`float`: Placement elevation of the context.

        This parameter only applies if the context does not have 3D placement enabled.
        If the context has 3D placement enabled, ``None`` is returned because the
        placement of the context is dictated by the underlying 3D transformation.
        Otherwise, the placement elevation of the context is returned.

        This property is read-only.
        """
        return (
            self.__stub.GetPlacementElevation(self.msg).value
            if not self.get_is_3d_placement(False)
            else None
        )

    def get_is_3d_placement(self, local):
        """Determine if the context has 3D placement enabled.

        Parameters
        ----------
        local : bool
            Whether 3D placement is enabled only in the local :class:`context <LayoutInstanceContext>`.
            If ``False``, a check is run to see if 3D placement is enabled anywhere higher in the
            hierarchy.

        Returns
        -------
        bool
            ``True`` if 3D placement is enabled in the local context.
        """
        return self.__stub.Is3DPlacement(bool_property_message(self, local)).value



"""Rectangle."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.geometry.polygon_data import PolygonData
    from ansys.edb.core.typing import NetLike, LayerLike, ValueLike

from enum import Enum

from ansys.api.edb.v1 import rectangle_pb2, rectangle_pb2_grpc

from ansys.edb.core.inner import messages, parser
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class RectangleRepresentationType(Enum):
    """Provides an enum representing rectangle types."""

    INVALID_RECT_TYPE = rectangle_pb2.INVALID_RECT_TYPE
    CENTER_WIDTH_HEIGHT = rectangle_pb2.CENTER_WIDTH_HEIGHT
    LOWER_LEFT_UPPER_RIGHT = rectangle_pb2.LOWER_LEFT_UPPER_RIGHT


class Rectangle(Primitive):
    """Represents a rectangle object."""

    __stub: rectangle_pb2_grpc.RectangleServiceStub = StubAccessor(StubType.rectangle)

    @classmethod
    def create(
        cls,
        layout: Layout,
        layer: LayerLike,
        net: NetLike | None,
        rep_type: RectangleRepresentationType,
        param1: ValueLike,
        param2: ValueLike,
        param3: ValueLike,
        param4: ValueLike,
        corner_rad: ValueLike,
        rotation: ValueLike,
    ) -> Rectangle:
        """Create a rectangle.

        Parameters
        ----------
        layout : .Layout
            Layout to create the rectangle in.
        layer : :term:`LayerLike`
            Layer the rectangle is to created on.
        net : :term:`NetLike` or None
            Net the rectangle is to have.
        rep_type : .RectangleRepresentationType
            Type that defines the meaning of the given parameters.
        param1 : :term:`ValueLike`
            X value of the lower-left point or center point.
        param2 : :term:`ValueLike`
            Y value of the lower-left point or center point.
        param3 : :term:`ValueLike`
            X value of the upper-right point or width.
        param4 : :term:`ValueLike`
            Y value of the upper-right point or height.
        corner_rad : :term:`ValueLike`
            Corner radius.
        rotation : :term:`ValueLike`
            Rotation.

        Returns
        -------
        Rectangle
            Rectangle created.
        """
        return Rectangle(
            cls.__stub.Create(
                rectangle_pb2.RectangleCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    net=messages.net_ref_message(net),
                    representation_type=rep_type.value,
                    parameter1=messages.value_message(param1),
                    parameter2=messages.value_message(param2),
                    parameter3=messages.value_message(param3),
                    parameter4=messages.value_message(param4),
                    corner_radius=messages.value_message(corner_rad),
                    rotation=messages.value_message(rotation),
                )
            )
        )

    def get_parameters(
        self,
    ) -> tuple[RectangleRepresentationType, Value, Value, Value, Value, Value, Value]:
        """Get coordinate parameters.

        Returns
        -------
        tuple of (.RectangleRepresentationType, .Value, .Value, .Value, .Value, .Value, .Value)

            Returns a tuple in this format:

            **(representation_type, parameter1, parameter2, parameter3, parameter4, corner_radius, rotation)**

            **representation_type** : Type that defines given parameters meaning.

            **parameter1** : X value of lower left point or center point.

            **parameter2** : Y value of lower left point or center point.

            **parameter3** : X value of upper right point or width.

            **parameter4** : Y value of upper right point or height.

            **corner_radius** : Corner radius.

            **rotation** : Rotation.
        """
        rect_param_msg = self.__stub.GetParameters(self.msg)
        return (
            RectangleRepresentationType(rect_param_msg.representation_type),
            Value(rect_param_msg.parameter1),
            Value(rect_param_msg.parameter2),
            Value(rect_param_msg.parameter3),
            Value(rect_param_msg.parameter4),
            Value(rect_param_msg.corner_radius),
            Value(rect_param_msg.rotation),
        )

    def set_parameters(
        self,
        rep_type: RectangleRepresentationType,
        param1: ValueLike,
        param2: ValueLike,
        param3: ValueLike,
        param4: ValueLike,
        corner_rad: ValueLike,
        rotation: ValueLike,
    ):
        """Set coordinate parameters.

        Parameters
        ----------
        rep_type : .RectangleRepresentationType
            Type that defines the meaning of the given parameters.
        param1 : :term:`ValueLike`
            X value of the lower-left point or center point.
        param2 : :term:`ValueLike`
            Y value of the lower-left point or center point.
        param3 : :term:`ValueLike`
            X value of the upper-right point or width.
        param4 : :term:`ValueLike`
            Y value of the upper-right point or height.
        corner_rad : :term:`ValueLike`
            Corner radius.
        rotation : :term:`ValueLike`
            Rotation.
        """
        self.__stub.SetParameters(
            rectangle_pb2.SetRectangleParametersMessage(
                target=self.msg,
                parameters=rectangle_pb2.RectangleParametersMessage(
                    representation_type=rep_type.value,
                    parameter1=messages.value_message(param1),
                    parameter2=messages.value_message(param2),
                    parameter3=messages.value_message(param3),
                    parameter4=messages.value_message(param4),
                    corner_radius=messages.value_message(corner_rad),
                    rotation=messages.value_message(rotation),
                ),
            )
        )

    @property
    def can_be_zone_primitive(self) -> bool:
        """:obj:`bool`: Flag indicating if the rectangle can be a zone.

        This property is read-only.
        """
        return True

    @property
    @parser.to_polygon_data
    def polygon_data(self) -> PolygonData:
        """:class:`.PolygonData`: \
        Polygon data object of the rectangle.

        This property is read-only.
        """
        return self.__stub.GetPolygonData(self.msg)

    @classmethod
    @parser.to_polygon_data
    def render(
        cls,
        rep_type: RectangleRepresentationType,
        x_lower_left_or_center_x: ValueLike,
        y_lower_left_or_center_y: ValueLike,
        x_upper_right_or_width: ValueLike,
        y_upper_right_or_height: ValueLike,
        corner_radius: ValueLike,
        rotation: ValueLike,
        is_hole: bool = False,
    ) -> PolygonData:
        """Get the polygon data of a rectangle.

        Parameters
        ----------
        rep_type : .RectangleRepresentationType
            Type that defines the meaning of the given parameters.
        x_lower_left_or_center_x : :term:`ValueLike`
            X value of the lower-left point or center point.
        y_lower_left_or_center_y : :term:`ValueLike`
            Y value of the lower-left point or center point.
        x_upper_right_or_width : :term:`ValueLike`
            X value of the upper-right point or width.
        y_upper_right_or_height : :term:`ValueLike`
            Y value of the upper-right point or height.
        corner_radius : :term:`ValueLike`
            Corner radius.
        rotation : :term:`ValueLike`
            Rotation.
        is_hole : bool, default: False
            Whether the rectangle is hole.

        Returns
        -------
        :class:`.PolygonData`
            Polygon data object created.
        """
        if rep_type == RectangleRepresentationType.LOWER_LEFT_UPPER_RIGHT:
            width = x_upper_right_or_width - x_lower_left_or_center_x
            height = y_upper_right_or_height - y_lower_left_or_center_y
            center_x = x_lower_left_or_center_x + width / 2.0
            center_y = y_lower_left_or_center_y + height / 2.0
            polygon_data = cls.__stub.Render(
                rectangle_pb2.RectanglePolygonDataMessage(
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    width=messages.value_message(width),
                    height=messages.value_message(height),
                    corner_radius=messages.value_message(corner_radius),
                    rotation=messages.value_message(rotation),
                    ishole=is_hole,
                )
            )
        elif rep_type == RectangleRepresentationType.CENTER_WIDTH_HEIGHT:
            polygon_data = cls.__stub.Render(
                rectangle_pb2.RectanglePolygonDataMessage(
                    center_x=messages.value_message(x_lower_left_or_center_x),
                    center_y=messages.value_message(y_lower_left_or_center_y),
                    width=messages.value_message(x_upper_right_or_width),
                    height=messages.value_message(y_upper_right_or_height),
                    corner_radius=messages.value_message(corner_radius),
                    rotation=messages.value_message(rotation),
                    ishole=is_hole,
                )
            )
        else:
            polygon_data = None
        return polygon_data



"""SIWave simulation setup."""

from ansys.edb.core.simulation_setup.simulation_setup import SimulationSetup, SimulationSetupType
from ansys.edb.core.simulation_setup.siwave_simulation_settings import SIWaveSimulationSettings


class SIWaveSimulationSetup(SimulationSetup):
    """Represents SIWave simulation setup data."""

    @classmethod
    def create(cls, cell, name):
        """Create a SIWave simulationsetup.

        Parameters
        ----------
        cell : :class:`.Cell`
            Cell to create simulation setup in.
        name : str
            Name of the simulation setup.

        Returns
        -------
        SIWaveSimulationSetup
            Simulation setup created.
        """
        return super()._create(cell, name, SimulationSetupType.SI_WAVE)

    @property
    def settings(self):
        """:class:`.SIWaveSimulationSettings`: Simulation settings of the SIWave simulation setup."""
        return SIWaveSimulationSettings(self)



"""Layout."""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

if TYPE_CHECKING:
    from ansys.edb.core.geometry.polygon_data import ExtentType, PolygonData
    from ansys.edb.core.layout.cell import Cell
    from ansys.edb.core.net.net import Net
    from ansys.edb.core.typing import LayerLike, ValueLike
    from ansys.edb.core.utility.value import Value

    LayerListLike = Union[LayerLike, List[LayerLike]]
    from src.ansys.edb.core.hierarchy.cell_instance import CellInstance
    from src.ansys.edb.core.hierarchy.group import Group
    from src.ansys.edb.core.hierarchy.pin_group import PinGroup
    from src.ansys.edb.core.layout.voltage_regulator import VoltageRegulator
    from src.ansys.edb.core.layout_instance.layout_instance import LayoutInstance
    from src.ansys.edb.core.net.differential_pair import DifferentialPair
    from src.ansys.edb.core.net.extended_net import ExtendedNet
    from src.ansys.edb.core.net.net_class import NetClass
    from src.ansys.edb.core.primitive.padstack_instance import PadstackInstance
    from src.ansys.edb.core.terminal.terminal import Terminal

from ansys.api.edb.v1 import layout_pb2
from ansys.api.edb.v1.edb_messages_pb2 import EDBObjMessage
from ansys.api.edb.v1.layout_pb2_grpc import LayoutServiceStub

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import ObjBase, messages, parser, utils, variable_server
from ansys.edb.core.layer.layer_collection import LayerCollection
from ansys.edb.core.layout.mcad_model import McadModel
from ansys.edb.core.layout_instance import layout_instance
from ansys.edb.core.primitive.board_bend_def import BoardBendDef
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType


def _geometry_simplifications_settings_msg(layout, layer, tol):
    """Create a GeometrySimplificationSettingsMessage."""
    return layout_pb2.GeometrySimplificationSettingsMessage(
        layout_layers=messages.layer_refs_property_message(layout, layer),
        tolerance=messages.value_message(tol),
    )


def _geometry_simplifications_settings_with_option_msg(layout, layer, tol, option):
    """Create a GeometrySimplificationSettingsWithOptionMessage."""
    return layout_pb2.GeometrySimplificationSettingsWithOptionMessage(
        geom_simplification_settings=_geometry_simplifications_settings_msg(layout, layer, tol),
        option=option,
    )


def _via_simplifications_settings_with_option_msg(
    layout, layer, tol, option, simplification_method
):
    """Create a ViaSimplificationSettingsMessage."""
    return layout_pb2.ViaSimplificationSettingsMessage(
        geom_simplification_settings_with_option=_geometry_simplifications_settings_with_option_msg(
            layout, layer, tol, option
        ),
        simplification_method=simplification_method,
    )


class Layout(ObjBase, variable_server.VariableServer):
    """Owns the geometry, nets, and stackup of an EDB design."""

    __stub: LayoutServiceStub = StubAccessor(StubType.layout)

    def __init__(self, msg: EDBObjMessage):
        """Initialize a new layout.

        Parameters
        ----------
        msg : :class:`.EDBObjMessage`
        """
        ObjBase.__init__(self, msg)
        variable_server.VariableServer.__init__(self, self)

    @property
    def cell(self) -> Cell:
        """:class:`.Cell`: Cell that owns the layout.

        This property is read-only.
        """
        from ansys.edb.core.layout.cell import Cell

        return Cell(self.__stub.GetCell(self.msg))

    @property
    def layer_collection(self) -> LayerCollection:
        """:class:`.LayerCollection`: Stackup of the layout."""
        return LayerCollection(self.__stub.GetLayerCollection(self.msg))

    @layer_collection.setter
    def layer_collection(self, layer_collection: LayerCollection):
        self.__stub.SetLayerCollection(
            layout_pb2.SetLayerCollectionMessage(
                layout=self.msg, layer_collection=layer_collection.msg
            )
        )

    def _get_items(self, lyt_obj_type_enum):
        """Get a list of layout objects."""
        return utils.query_lyt_object_collection(
            self, lyt_obj_type_enum, self.__stub.GetItems, self.__stub.StreamItems
        )

    @property
    def primitives(self) -> List[Primitive]:
        """:obj:`list` of :class:`.Primitive`: List of all primitives in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.PRIMITIVE)

    @property
    def padstack_instances(self) -> List[PadstackInstance]:
        """:obj:`list` of :class:`.PadstackInstance`: List of all padstack instances in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.PADSTACK_INSTANCE)

    @property
    def terminals(self) -> List[Terminal]:
        """:obj:`list` of :class:`.Terminal`: List of all terminals in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.TERMINAL)

    @property
    def cell_instances(self) -> List[CellInstance]:
        """:obj:`list` of :class:`.CellInstance`: List of all cell instances in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.CELL_INSTANCE)

    @property
    def nets(self) -> List[Net]:
        """:obj:`list` of :class:`.Net`: List of all nets in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.NET)

    @property
    def groups(self) -> List[Group]:
        """:obj:`list` of :class:`.Group`: List of all groups in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.GROUP)

    @property
    def net_classes(self) -> List[NetClass]:
        """:obj:`list` of :class:`.NetClass`: List of all net classes in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.NET_CLASS)

    @property
    def differential_pairs(self) -> List[DifferentialPair]:
        """:obj:`list` of :class:`.DifferentialPair`: List of all differential pairs in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.DIFFERENTIAL_PAIR)

    @property
    def pin_groups(self) -> List[PinGroup]:
        """:obj:`list` of :class:`.PinGroup`: List of all pin groups in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.PIN_GROUP)

    @property
    def voltage_regulators(self) -> List[VoltageRegulator]:
        """:obj:`list` of :class:`.VoltageRegulator`: List of all voltage regulators in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.VOLTAGE_REGULATOR)

    @property
    def extended_nets(self) -> List[ExtendedNet]:
        """:class:`list` of :class:`.ExtendedNet`: List of all extended nets in the layout.

        This property is read-only.
        """
        return self._get_items(LayoutObjType.EXTENDED_NET)

    @parser.to_polygon_data
    def expanded_extent(
        self,
        nets: List[Net],
        extent: ExtentType,
        expansion_factor: float,
        expansion_unitless: bool,
        use_round_corner: bool,
        num_increments: int,
    ) -> List[PolygonData]:
        """Get the expanded extent of the geometry on the specified nets.

        Parameters
        ----------
        nets : list of .Net
            List of nets containing the geometry.
        extent : .ExtentType
            Type of extent to be retrieved.
        expansion_factor : float
            Expansion factor applied to the extent. No expansion occurs if the value
            for this parameter is less than or equal to 0.
        expansion_unitless : bool
            When unitless, the distance by which the extent expands is the factor
            multiplied by the longer dimension (X or Y distance) of the bounding box of the extent.
        use_round_corner : bool
            Whether to use round corners or sharp corners. For round corners, this
            returns a bounding box if its area is within 10% of the rounded expansion's
            area.
        num_increments : int
            Number of iterations desired to reach the full expansion.

        Returns
        -------
        .PolygonData

        Notes
        -----
        This method returns the expansion of the contour, so any voids within expanded
        objects are ignored.
        """
        return self.__stub.GetExpandedExtentFromNets(
            messages.layout_expanded_extent_message(
                self,
                nets,
                extent,
                expansion_factor,
                expansion_unitless,
                use_round_corner,
                num_increments,
            )
        )

    def convert_primitives_to_vias(self, primitives: List[Primitive], is_pins: bool = False):
        """Convert a list of primitives into vias or pins.

        Parameters
        ----------
        primitives : list of .Primitive
            List of primitives to convert.
        is_pins : bool, default: False
            Flag indicating whether to convert the provided primitives
            to pins or vias. The default is ``False``, in which
            case the primitives will be converted to vias.
        """
        self.__stub.ConvertPrimitivesToVias(
            messages.layout_convert_p2v_message(self, primitives, is_pins)
        )

    @property
    def port_reference_terminals_connected(self) -> bool:
        """:obj:`bool`: Flag indicating if port reference terminals are connected.

        This property applies to lumped ports and circuit ports. It is ``True`` if
        port terminals are connected, ``False`` otherwise.

        This property is read-only.
        """
        return self.__stub.ArePortReferenceTerminalsConnected(self.msg).is_connected

    @property
    def zone_primitives(self) -> List[Primitive]:
        """:obj:`list` of :class:`.Primitive`: List of \
        all zone primitives in the layout.

        Zone primitives denote the outline of a :term:`zones <Zone>` in the layout.

        This property is read-only.
        """
        return [Primitive(msg).cast() for msg in self.__stub.GetZonePrimitives(self.msg).items]

    @property
    def fixed_zone_primitive(self) -> Primitive:
        """:class:`.Primitive`: Fixed :term:`zone <Zone>` primitive.

        The fixed zone primitive represents the zone that all bends are applied relative to.
        """
        msg = self.__stub.GetFixedZonePrimitive(self.msg)
        return None if msg is None else Primitive(msg).cast()

    @fixed_zone_primitive.setter
    def fixed_zone_primitive(self, value: Primitive):
        self.__stub.SetFixedZonePrimitives(messages.pointer_property_message(self, value))

    @property
    def board_bend_defs(self) -> List[BoardBendDef]:
        """:obj:`list` of :class:`.BoardBendDef`: List of all board bend definitions in the layout.

        This property is read-only.
        """
        return [BoardBendDef(msg) for msg in self.__stub.GetBoardBendDefs(self.msg)]

    def synchronize_bend_manager(self):
        """Synchronize the bend manager.

        Notes
        -----
        Most operations related to bends require this to be called first. If changes are made to the geometry
        in the design, it may be necessary to call this method again to ensure all bend related data is
        synchronized correctly.
        """
        self.__stub.SynchronizeBendManager(self.msg)

    @property
    def layout_instance(self) -> LayoutInstance:
        """:class:`.LayoutInstance`: Instance of the layout.

        This property is read-only.
        """
        return layout_instance.LayoutInstance(self.__stub.GetLayoutInstance(self.msg))

    def reconstruct_arcs(self, layer: LayerListLike, tolerance: Value):
        """Reconstruct arcs of polygons on a layer.

        Parameters
        ----------
        layer : :term:`LayerLike` or list of :term:`LayerLike`
            Layers to reconstruct arcs on.
        tolerance : .Value
            Tolerance.
        """
        self.__stub.ReconstructArcs(_geometry_simplifications_settings_msg(self, layer, tolerance))

    def unite_primitives(self, layer: LayerListLike):
        """Unite primitives on a layer.

        Parameters
        ----------
        layer : :term:`LayerLike` or list of :term:`LayerLike`
            Layers to unite primitives on.
        """
        self.__stub.UnitePrimitives(messages.layer_refs_property_message(self, layer))

    def create_stride(self, filename: str) -> McadModel:
        """Create a Stride model from an MCAD file.

        Parameters
        ----------
        filename : str
            Absolute path of the MCAD file.

        Returns
        -------
        .McadModel
           Stride model created.
        """
        return McadModel.create_stride(layout=self, filename=filename)

    def create_hfss(self, filename: str, design: str) -> McadModel:
        """Create an HFSS model from an MCAD file.

        Parameters
        ----------
        filename : str
            Absolute path of the MCAD file.
        design : str
            Design name.

        Returns
        -------
        .McadModel
            HFSS model created.
        """
        return McadModel.create_hfss(connectable=self, filename=filename, design=design)

    def create_3d_comp(self, filename: str) -> McadModel:
        """Create a 3D composite model from an MCAD file.

        Parameters
        ----------
        filename : str
            Absolute path of the MCAD file.

        Returns
        -------
        .McadModel
           3D composite model created.
        """
        return McadModel.create_3d_comp(layout=self, filename=filename)

    def group_vias(
        self,
        layer: LayerListLike,
        max_grouping_distance: ValueLike = "100um",
        persistent_vias: bool = False,
        group_by_proximity: bool = True,
        check_containment: bool = True,
    ):
        """Create via groups from the primitives on the specified layers.

        Parameters
        ----------
        layer : :term:`LayerLike` or list of :term:`LayerLike`
            Layers containing the primitives to be grouped.
        max_grouping_distance : :term:`ValueLike`
            Maximum distance between via primitives in a via group .
        persistent_vias : bool
            Whether to preserve primitives during via group creation. If ``False``
            primitives are deleted during via group creation.
        group_by_proximity : bool
            If ``True``, via primitives are grouped by proximity (relative position to each other).
            If ``False``, via primitives are grouped by range (any via primitives within the specified maximum
            grouping distance of each other are grouped)
        check_containment : bool
            If ``True``, the connectivity of via groups is checked and enforced to prevent
            short circuits in geometry connecting to the via group. If false, via primitives are
            grouped regardless of the connectivity of touching geometry.
        """
        self.__stub.GroupVias(
            layout_pb2.ViaGroupingSettingsMessage(
                via_simplification_settings=_via_simplifications_settings_with_option_msg(
                    self, layer, max_grouping_distance, check_containment, group_by_proximity
                ),
                persistent=persistent_vias,
            )
        )

    def snap_vias(
        self,
        layer: LayerListLike,
        via_snapping_tol: ValueLike = 3,
        prim_snapping_tol: ValueLike = "0.05um",
        snap_by_area_factor: bool = True,
        remove_dangling_vias: bool = True,
    ):
        """Snap vias on the specified layers to touching geometry.

        Parameters
        ----------
        layer : :term:`LayerLike` or list of :term:`LayerLike`
            Layer containing the vias to be snapped.
        via_snapping_tol : :term:`ValueLike`
            Tolerance for snapping vias. If snap_by_area_factor is ``True``, this
            value should not have a unit.
        prim_snapping_tol : :term:`ValueLike`
            Tolerance for snapping primitives.
        snap_by_area_factor : bool
            If ``True``, the via snapping tolerance is a factor of the surface area of the via.
            If ``False``, the via snapping tolerance is treated as an absolute distance.
        remove_dangling_vias : bool
            If ``True``, vias not connected to any geometry are removed.
        """
        self.__stub.SnapVias(
            layout_pb2.ViaSnappingSettingsMessage(
                via_simplification_settings=_via_simplifications_settings_with_option_msg(
                    self, layer, via_snapping_tol, remove_dangling_vias, snap_by_area_factor
                ),
                prim_snapping_tol=messages.value_message(prim_snapping_tol),
            )
        )

    def snap_primitives(
        self, layer: LayerListLike, tol: ValueLike = "0.05um", check_connectivity: bool = True
    ):
        """Snap primitives on the specified layer to touching geometry.

        Parameters
        ----------
        layer : str or list of str or .Layer or list of .Layer
            Layers containing the primitives to be snapped.
        tol : :term:`ValueLike`
            Tolerance for snapping primitives.
        check_connectivity : bool
            If ``True``, the connectivity of primitives is checked and enforced to prevent
            short circuits in geometry connecting to the primitives. If false, primitives are
            snapped regardless of the connectivity of touching geometry.
        """
        self.__stub.SnapPrimitives(
            _geometry_simplifications_settings_with_option_msg(self, layer, tol, check_connectivity)
        )

    def create_mesh_region(
        self,
        xy_exp: Value,
        pos_z_exp: Value,
        neg_z_exp: Value,
        use_active_nets: bool,
        incl_ref: bool,
        ext: ExtentType,
        num_x_partitions: int,
        num_y_partitions: int,
    ):
        """Designate a mesh region in a design and create partitions for meshing.

        Parameters
        ----------
        xy_exp : .Value
            Horizontal padding of the new mesh region (applied to both x and y directions).
        pos_z_exp : .Value
            Vertical padding above the new mesh region.
        neg_z_exp : .Value
            Vertical padding below the new mesh region.
        use_active_nets: bool
            ``True`` will create a new mesh region defined by the active nets in the design.
            ``False`` will create a new mesh region defined by the dielectric extents in the design.
        incl_ref : bool
            ``True`` will include bot positive nets and reference nets in the definition of the new mesh region.
            ``False`` will not include bot positive nets and reference nets.
        ext : .ExtentType
            Extent type of the new mesh region.
        num_x_partitions : int
            Number of partitions to create on x-axis in the new mesh region.
        num_y_partitions : int
            Number of partitions to create on y-axis in the new mesh region.
        """
        self.__stub.CreateMeshRegion(
            layout_pb2.CreateMeshRegionMessage(
                layout=self.msg,
                xy_exp=messages.value_message(xy_exp),
                pos_z_exp=messages.value_message(pos_z_exp),
                neg_z_exp=messages.value_message(neg_z_exp),
                use_active_nets=use_active_nets,
                incl_ref=incl_ref,
                ext=ext.value,
                num_x_partitions=num_x_partitions,
                num_y_partitions=num_y_partitions,
            )
        )

    def compress_primitives(self):
        """Compress :class:`primitives <.Primitive>` into \
        :class:`primitive instance collections <.PrimitiveInstanceCollection>`.

        Primitives whose only geometric difference is location will be compressed. \
        For example, a 4x4 grid of rectangles with the same width and height will be \
        compressed into one primitive instance collection.

        .. note::
           Only :class:`.Circle`, :class:`.Rectangle`, and :class:`.Polygon` primitives \
           are supported in primitive instance collections.
        """
        self.__stub.CompressPrimitives(self.msg)



"""Edge Terminal."""

from enum import Enum

import ansys.api.edb.v1.edge_term_pb2 as edge_term_pb2

from ansys.edb.core.geometry.arc_data import ArcData
from ansys.edb.core.inner import ObjBase, TypeField, messages, parser
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminal import Terminal, TerminalType


class EdgeType(Enum):
    """Provides an enum representing edge types."""

    PRIMITIVE = edge_term_pb2.PRIMITIVE_EDGE
    PADSTACK = edge_term_pb2.PAD_EDGE


class Edge(ObjBase):
    """Represents an edge."""

    __stub = StubAccessor(StubType.edge)
    type = TypeField(None)

    def cast(self):
        """Cast the base edge object to the correct subclass, if possible.

        Returns
        -------
        Edge
        """
        if self.is_null:
            return

        tp = self.type
        if tp == EdgeType.PRIMITIVE:
            return PrimitiveEdge(self.msg)
        if tp == EdgeType.PADSTACK:
            return PadEdge(self.msg)

    @classmethod
    def _create(cls, **params):
        return cls.__stub.Create(messages.edge_creation_message(cls.type.value, **params))

    @property
    def _type(self):
        return EdgeType(self.__stub.GetEdgeType(self.msg).t)

    @property
    def _params(self):
        res = self.__stub.GetParameters(self.msg)
        if self.type == EdgeType.PRIMITIVE:
            return res.primitive_params
        elif self.type == EdgeType.PADSTACK:
            return res.pad_params


class PadEdge(Edge):
    """Represents a padstack edge."""

    type = TypeField(EdgeType.PADSTACK)

    @classmethod
    def create(cls, padstack_instance, layer, arc):
        """Create a padstack edge.

        Parameters
        ----------
        padstack_instance : :class:`.PadstackInstance`
        layer : :class:`.Layer` or :obj:`str`
        arc : :class:`.ArcData`

        Returns
        -------
        PadEdge
        """
        return PadEdge(cls._create(padstack_instance=padstack_instance, layer=layer, arc=arc))

    @property
    def padstack_instance(self):
        """:class:`.PadstackInstance`: Padstack instance that the edge is on."""
        from ansys.edb.core.primitive.padstack_instance import PadstackInstance

        return PadstackInstance(self._params.padstack_instance)

    @property
    def layer(self):
        """:class:`.Layer`: Layer that the edge is on."""
        return Layer(self._params.layer.id).cast()

    @property
    def arc(self):
        """:class:`.ArcData`: Arc of the edge."""
        return ArcData(self._params.arc)


class PrimitiveEdge(Edge):
    """Represents a primitive edge."""

    type = TypeField(EdgeType.PRIMITIVE)

    @classmethod
    def create(cls, prim, point):
        """Create a primitive edge.

        Parameters
        ----------
        prim : :class:`.Primitive`
        point : :term:`Point2DLike`

        Returns
        -------
        PrimitiveEdge
        """
        return PrimitiveEdge(cls._create(primitive=prim, point=point))

    @property
    def primitive(self):
        """:class:`.Primitive`: Primitive of the terminal."""
        from ansys.edb.core.primitive import primitive

        return primitive.Primitive(self._params.primitive).cast()

    @property
    @parser.to_point_data
    def point(self):
        """:class:`.PointData`: Coordinates (x, y) of the terminal."""
        return self._params.point


class EdgeTerminal(Terminal):
    """Represents an edge terminal."""

    __stub = StubAccessor(StubType.edge_terminal)
    type = TypeField(TerminalType.EDGE)

    @classmethod
    def create(cls, layout, name, edges, net=None, is_ref=False):
        """Create an edge terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the edge terminal in.
        name : :obj:`str`
            Name of the edge terminal.
        edges : list of Edge
        net : :class:`.Net` or :obj:`str` or None
            Net reference. The default is ``None``.
        is_ref : :obj:`bool`, default: False
            Whether the edge terminal is a reference terminal.

        Returns
        -------
        EdgeTerminal
        """
        return EdgeTerminal(
            cls.__stub.Create(messages.edge_term_creation_message(layout, net, name, edges, is_ref))
        )

    @property
    def edges(self):
        """:obj:`list` of :class:`.Edge`: All edges on the terminal."""
        return [Edge(msg).cast() for msg in self.__stub.GetEdges(self.msg).items]

    @edges.setter
    def edges(self, edges):
        self.__stub.GetEdges(messages.edge_term_set_edges_message(self, edges))



"""Heat sink."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike

from enum import Enum

import ansys.api.edb.v1.package_def_pb2 as pb

from ansys.edb.core.utility import conversions


class HeatSinkFinOrientation(Enum):
    """Provides an enum representing bondwire types."""

    X_ORIENTED = pb.X_ORIENTED
    Y_ORIENTED = pb.Y_ORIENTED
    OTHER_ORIENTED = pb.OTHER_ORIENTED


class HeatSink:
    """Represents a heat sink, a three-dimensional modeling component \
    that can exchange radiation with other objects in the model.

    Attributes
    ----------
        fin_thickness : :term:`ValueLike`
            Fin thickness of the heat sink.
        fin_spacing : :term:`ValueLike`
            Fin spacing of the heat sink.
        fin_base_height : :term:`ValueLike`
            Base elevation of the the heat sink.
        fin_height : :term:`ValueLike`
            Fin height of the heat sink.
        fin_orientation : .HeatSinkFinOrientation
            Fin orientation of the heat sink if it is not set to the X axis.
    """

    def __init__(
        self,
        fin_thickness: ValueLike = 0,
        fin_spacing: ValueLike = 0,
        fin_base_height: ValueLike = 0,
        fin_height: ValueLike = 0,
        fin_orientation: HeatSinkFinOrientation = HeatSinkFinOrientation.X_ORIENTED,
    ):
        """Initialize a heat sink object using given values.

        Parameters
        ----------
        fin_thickness : :term:`ValueLike`
            Fin thickness of the heat sink.
        fin_spacing : :term:`ValueLike`
            Fin spacing of the heat sink.
        fin_base_height : :term:`ValueLike`
            Base elevation of the the heat sink.
        fin_height : :term:`ValueLike`
            Fin height of the heat sink.
        fin_orientation : .HeatSinkFinOrientation
            Fin orientation of the heat sink if it is not set to the X axis.
        """
        self.fin_thickness = conversions.to_value(fin_thickness)
        self.fin_spacing = conversions.to_value(fin_spacing)
        self.fin_base_height = conversions.to_value(fin_base_height)
        self.fin_height = conversions.to_value(fin_height)
        self.fin_orientation = fin_orientation



"""Base hierarchy model."""
from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import ModelServiceStub, StubAccessor, StubType


class Model(ObjBase):
    """Represents a base hierarchy model object."""

    __stub: ModelServiceStub = StubAccessor(StubType.model)

    def clone(self):
        """Clone a model.

        Returns
        -------
        Model cloned.
        """
        return self.__class__(self.__stub.Clone(messages.edb_obj_message(self)))



"""Value."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike, Union
    from ansys.edb.core.database import Database
    from ansys.edb.core.layout.cell import Cell
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.definition.component_def import ComponentDef


import math

from ansys.api.edb.v1 import value_pb2, value_pb2_grpc
from ansys.api.edb.v1.edb_messages_pb2 import ValueMessage

from ansys.edb.core import session
from ansys.edb.core.inner import messages
from ansys.edb.core.utility import conversions


class Value:
    r"""Represents a number or an expression.

    Attributes
    ----------
    val : :term:`ValueLike`
        Value assigned to the new value.
    _owner :    None, :class:`.Database`,
                :class:`.Cell`,
                :class:`.Layout`,
                :class:`.ComponentDef`

    Notes
    -----
    A value can be either a constant (such as ``1``, ``2.35``, ``"7+0.3i"``, and ``"23mm"``) or
    parametric (such as.``w1 + w2``).

    If the value is parametric, the ``_owner`` attribute must be set to the object that hosts the
    variables used. If the owner is :class:`.Cell` or
    :class:`.Layout`, the expression can reference both database variables
    and cell variables. A better way to create a parametric values is to call the
    ``obj_with_variables.create_value(str)`` method, which automatically sets the
    ``_owner`` parameter to the correct object.

    Values can be used in expressions with the following operators:

    .. list-table:: **Mathematical operators supported by values**
       :widths: 25 25 25
       :header-rows: 1

       * - Operator
         - Operation
         - Return Value
       * - \+
         - addition
         - Value
       * - \-
         - subtraction or negation
         - Value
       * - \*
         - multiplication
         - Value
       * - /
         - division
         - Value
       * - //
         - floor division
         - integer
       * - \*\*
         - power
         - Value
       * - ==
         - equality
         - bool
       * - \<
         - less than
         - bool
       * - >
         - greater than
         - bool

    The value is evaluated to a constant (if it is parametric) before applying the operators.
    """

    __stub: value_pb2_grpc.ValueServiceStub = session.StubAccessor(session.StubType.value)

    def __init__(self, val: ValueLike, _owner: Union[Database, Layout, Cell, ComponentDef] = None):
        """Initialize a value object."""
        self.msg = ValueMessage()
        if isinstance(val, ValueMessage):
            self.msg = val
        elif isinstance(val, Value):
            self.msg = val.msg
        elif isinstance(val, str):
            temp = value_pb2.ValueTextMessage(
                text=val, variable_owner=messages.edb_obj_message(_owner)
            )
            self.msg = self.__stub.CreateValue(temp)
        elif isinstance(val, float) or isinstance(val, int):
            self.msg.constant.real = val
            self.msg.constant.imag = 0
        elif isinstance(val, complex):
            self.msg.constant.real = val.real
            self.msg.constant.imag = val.imag
        else:
            raise TypeError(f"invalid value. Received '{val}'")

    def __str__(self):
        """Generate a readable string for the value.

        Returns
        -------
        str
        """
        if self.msg.text:
            return self.msg.text
        elif self.msg.constant.imag == 0:
            return str(self.msg.constant.real)
        else:
            return str(complex(self.msg.constant.real, self.msg.constant.imag))

    def __eq__(self, other: Value):
        """Compare this value to another value to determine if they are equivalent.

        Parameters
        ----------
        other : .Value
            Value to compare.

        Returns
        -------
        bool
            ``True`` if the two values are equivalent, ``False`` otherwise.
        """
        return self.equals(other)

    def __add__(self, other: ValueLike):
        """Add this value and another value.

        Parameters
        ----------
        other : :term:`ValueLike`
            Other value to add.


        Returns
        -------
        .Value
            Constant value wrapping either a real or complex number.
        """
        other = conversions.to_value(other)
        return self.__class__(self.value + other.value)

    def __sub__(self, other: ValueLike):
        """Subtract this value from another value.

        Parameters
        ----------
        other : :term:`ValueLike`
           Other value to subtract.

        Returns
        -------
        .Value
            Constant Value wrapping either a real or complex number.
        """
        other = conversions.to_value(other)
        return self.__class__(self.value - other.value)

    def __mul__(self, other: ValueLike):
        """Multiply this value and another value.

        Parameters
        ----------
        other : :term:`ValueLike`
            Other value to multiply.

        Returns
        -------
        .Value
            Constant Value wrapping either a real or complex number.
        """
        other = conversions.to_value(other)
        return self.__class__(self.value * other.value)

    def __truediv__(self, other: ValueLike):
        """Perform floating-point division of this value and another value.

        Parameters
        ----------
        other : :term:`ValueLike`
             Other value for the floating-point division.

        Returns
        -------
        .Value
            Constant value wrapping either a real or complex number.
        """
        other = conversions.to_value(other)
        return self.__class__(self.value / other.value)

    def __floordiv__(self, other: ValueLike):
        """Divide this value by another value and return its floor (integer part).

        Parameters
        ----------
        other : :term:`ValueLike`
            Other value for the division.

        Returns
        -------
        .Value
            Constant value wrapping an integer.
        """
        other = conversions.to_value(other)
        return self.__class__(self.value // other.value)

    def __pow__(self, power: Union[int, float], modulo=None):
        """Raise a value to the power of another value.

        Parameters
        ----------
        power : int, float
            Exponent to apply to the value.

        Returns
        -------
        .Value
            Constant value wrapping either a real or complex number.
        """
        return self.__class__(self.value**power)

    def __neg__(self):
        """Flip the sign.

        Returns
        -------
        .Value
            Constant value wrapping either a real or complex number.
        """
        return self.__class__(-self.value)

    def __gt__(self, other: ValueLike):
        """Compare this value to another to see if this value is greater.

        Parameters
        ----------
        other : :term:`ValueLike`
            Other value to compare to.

        Returns
        -------
        bool
            ``True`` if this value is greater than the other value.
        """
        return self.value > other.value

    def __lt__(self, other: ValueLike):
        """Compare this value to another to see if this value is less.

        Parameters
        ----------
        other : :term:`ValueLike`
            Other value to compare to.

        Returns
        -------
        bool
            ``True`` if this value is less than the other value.
        """
        return self.value < other.value

    def equals(self, other: ValueLike, tolerance: float = 1e-9):
        """Check if this value and other value are equivalent when evaluated.

        Parameters
        ----------
        other : :term:`ValueLike`
            Other value to compare to.
        tolerance : float, default: 1e-9
            Tolerance.

        Returns
        -------
        bool
            `True`` if this value and the other value are equivalent when evaluated.
        """
        try:
            other = conversions.to_value(other)
            diff = self.value - other.value

            if type(diff) == complex:
                return math.fabs(diff.real) <= tolerance and math.fabs(diff.imag) <= tolerance
            else:
                return math.fabs(diff) <= tolerance
        except TypeError:
            return False

    @property
    def is_parametric(self):
        """:obj:`bool`: Flag indicating if the value is parametric (dependent on variables)."""
        return not self.msg.HasField("constant")

    @property
    def is_complex(self):
        """:obj:`bool`: Flag indicating if the the value is a complex number (has a non-zero imaginary part)."""
        return type(self.value) == complex

    @property
    def double(self):
        """:obj:`float`: Float from the value object.

        If the number is complex, this returns the real part.
        """
        evaluated = self.value
        return evaluated.real if type(evaluated) == complex else evaluated

    @property
    def complex(self):
        """:obj:`complex`: Complex value from the value object.

        If the number is real, the imaginary part is ``0``.
        """
        return complex(self.value)

    @property
    def value(self):
        """:obj:`complex`: Evaluation to a constant and return as a float or complex."""
        if self.is_parametric:
            evaluated = self.__stub.GetComplex(
                value_pb2.ValueTextMessage(
                    text=self.msg.text, variable_owner=self.msg.variable_owner
                )
            )
        else:
            evaluated = self.msg.constant

        if evaluated.imag == 0:
            return evaluated.real
        else:
            return complex(evaluated.real, evaluated.imag)

    @property
    def sqrt(self):
        """:obj:`float`: Square root of this value as a constant value."""
        return self**0.5



"""Point Terminal."""

from ansys.edb.core.inner import TypeField, messages, parser
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminal import Terminal, TerminalType


class PointTerminal(Terminal):
    """Represents a point terminal object."""

    __stub = StubAccessor(StubType.point_terminal)
    type = TypeField(TerminalType.POINT)

    @classmethod
    def create(cls, layout, net, layer, name, point):
        """Create a point terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the point terminal in.
        net : :class:`.Net` or :obj:`str` or None
            Net.
        layer : :class:`.Layer` or :obj:`str`
            Layer to place the point terminal on.
        name : :obj:`str`
            Name of the point terminal.
        point : :term:`Point2DLike`
            Type of the point terminal.

        Returns
        -------
        PointTerminal
        """
        return PointTerminal(
            cls.__stub.Create(messages.point_term_creation_message(layout, net, layer, name, point))
        )

    @property
    def params(self):
        """:class:`.Layer`, :class:`.PointData`: Layer that the point terminal is placed on and \
        the (x, y) coordinates."""
        res = self.__stub.GetParameters(self.msg)
        point = parser.to_point_data(res.point)
        layer = Layer(res.layer.id).cast()
        return layer, point

    @params.setter
    def params(self, params):
        self.__stub.SetParameters(messages.point_term_set_params_message(self, *params))

    @property
    def layer(self):
        """:class:`.Layer`: Layer that the point terminal is placed on."""
        return self.params[0]

    @layer.setter
    def layer(self, value):
        self.params = (value, self.point)

    @property
    def point(self):
        """:class:`.PointData`: Coordinates (x, y) of the point terminal."""
        return self.params[1]

    @point.setter
    def point(self, value):
        self.params = (self.layer, value)



"""Primitive Instance Collection."""
from __future__ import annotations

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from ansys.edb.core.geometry.polygon_data import PolygonData
    from ansys.edb.core.geometry.point_data import PointData
    from ansys.edb.core.typing import PointLike, LayerLike, NetLike
    from ansys.edb.core.layout.layout import Layout

from ansys.api.edb.v1.primitive_instance_collection_pb2 import (
    PrimitiveInstanceCollectionDataMessage,
)
from ansys.api.edb.v1.primitive_instance_collection_pb2_grpc import (
    PrimitiveInstanceCollectionServiceStub,
)

from ansys.edb.core.inner.layout_obj import LayoutObj
import ansys.edb.core.inner.messages as messages
from ansys.edb.core.inner.parser import msg_to_point_data, msg_to_polygon_data, to_polygon_data
from ansys.edb.core.inner.utils import client_stream_iterator, stream_items_from_server
from ansys.edb.core.session import StubAccessor, StubType


class PrimitiveInstanceCollection(LayoutObj):
    """Efficiently represents large quantities of geometry as \
    numerous instantiations of the same geometry at different locations."""

    __stub: PrimitiveInstanceCollectionServiceStub = StubAccessor(
        StubType.primitive_instance_collection
    )

    @staticmethod
    def _point_request_iterator(points, starting_chunk):
        chunk_entry_creator = lambda point: messages.point_message(point)
        chunk_entries_getter = lambda chunk: chunk.points.points
        return client_stream_iterator(
            points,
            PrimitiveInstanceCollectionDataMessage,
            chunk_entry_creator,
            chunk_entries_getter,
            8000,
            starting_chunk,
        )

    @classmethod
    def create(
        cls,
        layout: Layout,
        net: NetLike,
        layer: LayerLike,
        geometry: PolygonData,
        positions: List[PointLike],
    ) -> PrimitiveInstanceCollection:
        """Create a primitive instance collection containing the specified geometry instantiated \
        at the provided locations. All geometry will be created on the specified layer and net.

        Parameters
        ----------
        layout : .Layout
            Layout to create the primitive instance collection in.
        net : :term:`NetLike`
            Net that the primitive instance collection geometry will exist on.
        layer : :term:`LayerLike`
            Layer that the primitive instance collection geometry will exist on.
        geometry : .PolygonData
            The geometry that will be instantiated.
        positions : list of :term:`Point2DLike`
            The points to instantiate the geometry at.

        Returns
        -------
        .PrimitiveInstanceCollection
        """
        chunk = PrimitiveInstanceCollectionDataMessage(
            lyt_or_prim_inst_col=layout.msg,
            net=messages.net_ref_message(net),
            layer=messages.layer_ref_message(layer),
            geometry=messages.polygon_data_message(geometry),
        )
        return PrimitiveInstanceCollection(
            cls.__stub.Create(cls._point_request_iterator(positions, chunk))
        )

    @property
    @to_polygon_data
    def geometry(self) -> PolygonData:
        """:class:`.PolygonData`: The geometry that the primitive instance collection instantiates."""
        return self.__stub.GetGeometry(self.msg)

    @geometry.setter
    def geometry(self, geometry: PolygonData):
        self.__stub.SetGeometry(messages.polygon_data_property_message(self, geometry))

    @property
    def positions(self) -> List[PointData]:
        """:obj:`list` of :class:`.PointData`: The positions geometry is instantiated at."""
        return stream_items_from_server(
            msg_to_point_data, self.__stub.GetPositions(self.msg), "points"
        )

    @positions.setter
    def positions(self, positions: List[PointData]):
        chunk = PrimitiveInstanceCollectionDataMessage(lyt_or_prim_inst_col=self.msg)
        self.__stub.SetPositions(self._point_request_iterator(positions, chunk))

    @property
    def instantiated_geometry(self) -> List[PolygonData]:
        """:obj:`list` of :class:`.PolygonData`: The geometry instantiated at each location in \
        the primitive instance collection.

        This property is read-only.
        """
        return stream_items_from_server(
            msg_to_polygon_data, self.__stub.GetInstantiatedGeometry(self.msg), "polygons"
        )

    def decompose(self):
        """Decompose into individual :class:`primitives <.Primitive>`. A \
        :class:`primitive <.Primitive>` will be created for each geometry instantiation."""
        return self.__stub.Decompose(self.msg)



"""Net."""

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import layout_obj, messages, utils
from ansys.edb.core.session import NetServiceStub, StubAccessor, StubType


class Net(layout_obj.LayoutObj):
    """Represents a net."""

    layout_obj_type = LayoutObjType.NET
    no_net_name = "<NO-NET>"
    __stub: NetServiceStub = StubAccessor(StubType.net)

    def _layout_objs(self, obj_type):
        """Get layout objects on a net."""
        return utils.query_lyt_object_collection(
            self, obj_type, self.__stub.GetLayoutObjects, self.__stub.StreamLayoutObjects
        )

    @classmethod
    def create(cls, layout, name):
        """Create a net.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the net in.
        name : str
            Name of the net.

        Returns
        -------
        Net
            Net created.
        """
        return Net(cls.__stub.Create(messages.string_property_message(layout, name)))

    @classmethod
    def find_by_name(cls, layout, name):
        """Find a net by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the net.
        name : str
            Name of net.

        Returns
        -------
        Net
            Net found. Check the :obj:`is_null <.Net.is_null>` property
            of the returned net to see if it exists.
        """
        return Net(cls.__stub.FindByName(messages.string_property_message(layout, name)))

    @property
    def name(self):
        """:class:`str`: Name of the net."""
        return self.no_net_name if self.is_no_net else self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, value):
        self.__stub.SetName(messages.string_property_message(self, value))

    @property
    def is_power_ground(self):
        """:class:`bool`: Flag indicating if the net belongs to a power/ground :class:`.NetClass` instance.

        This property is read-only.
        """
        return self.__stub.GetIsPowerGround(self.msg).value

    @is_power_ground.setter
    def is_power_ground(self, value):
        self.__stub.SetIsPowerGround(messages.bool_property_message(self, value))

    @property
    def primitives(self):
        r""":obj:`list` of :class:`.Primitive`: All primitives on the net.

        This property is read-only.
        """
        return self._layout_objs(LayoutObjType.PRIMITIVE)

    @property
    def padstack_instances(self):
        """:obj:`list` of :class:`.PadstackInstance`: All padstack instances on the net \
        object instance."""
        return self._layout_objs(LayoutObjType.PADSTACK_INSTANCE)

    @property
    def terminals(self):
        """:obj:`list` of :class:`.Terminal`: All terminal instances on the \
            net object instance."""
        return self._layout_objs(LayoutObjType.TERMINAL)

    @property
    def terminal_instances(self):
        """:obj:`list` of :class:`.Layer`: All terminal instances on the net object instance."""
        return self._layout_objs(LayoutObjType.TERMINAL_INSTANCE)

    @property
    def net_classes(self):
        r""":obj:`list` of :class:`.NetClass`: All net classes on the net.

        This property is read-only.
        """
        return self._layout_objs(LayoutObjType.NET_CLASS)

    @property
    def extended_net(self):
        """:class:`.ExtendedNet` or :class:`None`: Extended net that the net belongs to.

        :class:`None` means that the net does not belong to an extended net.

        This property is read-only.
        """
        en = self._layout_objs(LayoutObjType.EXTENDED_NET)
        if not en:
            return None
        en = en[0]
        return None if en.is_null else en

    @property
    def is_no_net(self):
        """:obj:`bool` Flag indicating if this the "no net" (empty) net."""
        return self.is_null



"""HFSS simulation settings."""

from enum import Enum

import ansys.api.edb.v1.hfss_simulation_settings_pb2 as pb

from ansys.edb.core.inner import messages, parser
from ansys.edb.core.session import (
    DCRSettingsServiceStub,
    HFSSAdvancedMeshingSettingsServiceStub,
    HFSSAdvancedSettingsServiceStub,
    HFSSGeneralSettingsServiceStub,
    HFSSOptionsSettingsServiceStub,
    HFSSSolverSettingsServiceStub,
    StubAccessor,
    StubType,
)
from ansys.edb.core.simulation_setup.simulation_settings import (
    AdvancedMeshingSettings,
    AdvancedSettings,
    SettingsOptions,
    SimulationSettings,
    SimulationSettingsBase,
    SolverSettings,
)


class BasisFunctionOrder(Enum):
    """Provides an enum representing basis function order types."""

    ZERO_ORDER = pb.ZERO_ORDER
    FIRST_ORDER = pb.FIRST_ORDER
    SECOND_ORDER = pb.SECOND_ORDER
    MIXED_ORDER = pb.MIXED_ORDER


class SolverType(Enum):
    """Provides an enum representing HFSS solver types."""

    AUTO_SOLVER = pb.AUTO_SOLVER
    DIRECT_SOLVER = pb.DIRECT_SOLVER
    ITERATIVE_SOLVER = pb.ITERATIVE_SOLVER
    NUM_SOLVER_TYPES = pb.NUM_SOLVER_TYPES


class AdaptType(Enum):
    """Provides an enum representing HFSS adaptive solution types."""

    SINGLE = pb.SINGLE
    MULTI_FREQUENCIES = pb.MULTI_FREQUENCIES
    BROADBAND = pb.BROADBAND
    NUM_ADAPT_TYPE = pb.NUM_ADAPT_TYPE


class HFSSSimulationSettings(SimulationSettings):
    """Represents HFSS simulation settings."""

    @property
    def general(self):
        """:class:`.HFSSGeneralSettings`: General settings for HFSS simulations."""
        return HFSSGeneralSettings(self._sim_setup)

    @property
    def options(self):
        """:class:`.HFSSSettingsOptions`: HFSS simulation setting options."""
        return HFSSSettingsOptions(self._sim_setup)

    @property
    def advanced(self):
        """:class:`.HFSSAdvancedSettings`: Advanced settings for HFSS simulations."""
        return HFSSAdvancedSettings(self._sim_setup)

    @property
    def advanced_meshing(self):
        """:class:`.HFSSAdvancedMeshingSettings`: Advanced meshing settings for HFSS simulations."""
        return HFSSAdvancedMeshingSettings(self._sim_setup)

    @property
    def solver(self):
        """:class:`.HFSSSolverSettings`: Solver settings for HFSS simulations."""
        return HFSSSolverSettings(self._sim_setup)

    @property
    def dcr(self):
        """:class:`.HFSSDCRSettings`: DCR settings for HFSS simulations."""
        return HFSSDCRSettings(self._sim_setup)


class HFSSGeneralSettings(SimulationSettingsBase):
    """Represents general settings for HFSS simulations."""

    __stub: HFSSGeneralSettingsServiceStub = StubAccessor(StubType.hfss_general_sim_settings)

    @property
    def single_frequency_adaptive_solution(self):
        """.SingleFrequencyAdaptiveSolution`: Settings for a single frequency adaptive solution."""
        return parser.to_single_frequency_adaptive_solution(
            self.__stub.GetSingleFrequencyAdaptiveSolution(self.msg)
        )

    @single_frequency_adaptive_solution.setter
    def single_frequency_adaptive_solution(self, single_frequency_adaptive_solution):
        self.__stub.SetSingleFrequencyAdaptiveSolution(
            pb.SingleFrequencyAdaptiveSolutionPropertyMessage(
                target=self.msg,
                adaptive_frequency=messages.single_frequency_adaptive_solution_msg(
                    single_frequency_adaptive_solution
                ),
            )
        )

    @property
    def multi_frequency_adaptive_solution(self):
        """:class:`.MultiFrequencyAdaptiveSolution`: Settings for a multi-frequency adaptive solution."""
        return parser.to_multi_frequency_adaptive_solution(
            self.__stub.GetMultiFrequencyAdaptiveSolution(self.msg)
        )

    @multi_frequency_adaptive_solution.setter
    def multi_frequency_adaptive_solution(self, multi_frequency_adaptive_solution):
        self.__stub.SetMultiFrequencyAdaptiveSolution(
            pb.MultiFrequencyAdaptiveSolutionPropertyMessage(
                target=self.msg,
                adaptive_frequency=messages.multi_frequency_adaptive_solution_msg(
                    multi_frequency_adaptive_solution
                ),
            )
        )

    @property
    def broadband_adaptive_solution(self):
        """:class:`.BroadbandAdaptiveSolution`: Settings for a broadband adaptive solution."""
        return parser.to_broadband_adaptive_solution(
            self.__stub.GetBroadbandFrequencyAdaptiveSolution(self.msg)
        )

    @broadband_adaptive_solution.setter
    def broadband_adaptive_solution(self, broadband_adaptive_solution):
        self.__stub.SetBroadbandFrequencyAdaptiveSolution(
            pb.BroadbandFrequencyAdaptiveSolutionPropertyMessage(
                target=self.msg,
                adaptive_frequency=messages.broadband_solution_msg(broadband_adaptive_solution),
            )
        )

    @property
    def adaptive_solution_type(self):
        """:class:`AdaptType`: Adaptive solution type that is set for the simulation."""
        return AdaptType(self.__stub.GetAdaptType(self.msg).adapt_type)

    @adaptive_solution_type.setter
    def adaptive_solution_type(self, adaptive_solution_type):
        self.__stub.SetAdaptType(
            pb.AdaptTypePropertyMessage(target=self.msg, adapt_type=adaptive_solution_type.value)
        )

    @property
    def save_fields(self):
        """:obj:`bool`: Flag indicating if field data is to be saved during the simulation."""
        return self.__stub.GetSaveFieldsFlag(self.msg).value

    @save_fields.setter
    def save_fields(self, save_fields):
        self.__stub.SetSaveFieldsFlag(messages.bool_property_message(self, save_fields))

    @property
    def save_rad_fields_only(self):
        """:obj:`bool`: Flag indicating if only radiated field data is to be saved during the simulation."""
        return self.__stub.GetSaveRadFieldsOnlyFlag(self.msg).value

    @save_rad_fields_only.setter
    def save_rad_fields_only(self, save_rad_fields_only):
        self.__stub.SetSaveRadFieldsOnlyFlag(
            messages.bool_property_message(self, save_rad_fields_only)
        )

    @property
    def use_mesh_region(self):
        """:obj:`bool`: Flag indicating if mesh regions are used."""
        return self.__stub.GetUseMeshRegion(self.msg).value

    @use_mesh_region.setter
    def use_mesh_region(self, use_mesh_region):
        self.__stub.SetUseMeshRegion(messages.bool_property_message(self, use_mesh_region))

    @property
    def mesh_region_name(self):
        """:obj:`str`: Name of the mesh region to use."""
        return self.__stub.GetMeshRegionName(self.msg).value

    @mesh_region_name.setter
    def mesh_region_name(self, mesh_region_name):
        self.__stub.SetMeshRegionName(messages.string_property_message(self, mesh_region_name))

    @property
    def use_parallel_refinement(self):
        """:obj:`bool`: Flag indicating if parallel refinement is used."""
        return self.__stub.GetUseParallelRefinement(self.msg).value

    @use_parallel_refinement.setter
    def use_parallel_refinement(self, use_parallel_refinement):
        self.__stub.SetUseParallelRefinement(
            messages.bool_property_message(self, use_parallel_refinement)
        )


class HFSSSettingsOptions(SettingsOptions):
    """Represents options for HFSS simulation settings."""

    __stub: HFSSOptionsSettingsServiceStub = StubAccessor(StubType.hfss_options_sim_settings)

    @property
    def use_max_refinement(self):
        """:obj:`bool`: Flag indicating if maximum refinement values are used during simulation."""
        return self.__stub.GetUseMaxRefinement(self.msg).value

    @use_max_refinement.setter
    def use_max_refinement(self, use_max_refinement):
        self.__stub.SetUseMaxRefinement(messages.bool_property_message(self, use_max_refinement))

    @property
    def max_refinement_per_pass(self):
        """:obj:`int`: Maximum mesh refinement per adaptive pass."""
        return self.__stub.GetMaxRefinementPerPass(self.msg).value

    @max_refinement_per_pass.setter
    def max_refinement_per_pass(self, max_refinement_per_pass):
        self.__stub.SetMaxRefinementPerPass(
            messages.uint64_property_message(self, max_refinement_per_pass)
        )

    @property
    def min_passes(self):
        """:obj:`int`: Minimum number of adaptive passes."""
        return self.__stub.GetMinPasses(self.msg).value

    @min_passes.setter
    def min_passes(self, min_refinement_passes):
        self.__stub.SetMinPasses(messages.uint64_property_message(self, min_refinement_passes))

    @property
    def min_converged_passes(self):
        """:obj:`int`: Minimum number of converged adaptive passes."""
        return self.__stub.GetMinConvergedPasses(self.msg).value

    @min_converged_passes.setter
    def min_converged_passes(self, min_refinement_passes):
        self.__stub.SetMinConvergedPasses(
            messages.uint64_property_message(self, min_refinement_passes)
        )

    @property
    def order_basis(self):
        """:class:`.BasisFunctionOrder`: Basis function order."""
        return BasisFunctionOrder(self.__stub.GetBasisFunctionOrder(self.msg).basis_function_order)

    @order_basis.setter
    def order_basis(self, order_basis):
        self.__stub.SetBasisFunctionOrder(
            pb.BasisFunctionOrderPropertyMessage(
                target=self.msg, basis_function_order=order_basis.value
            )
        )

    @property
    def solver_type(self):
        """:class:`.SolverType`: HFSS solver type."""
        return SolverType(self.__stub.GetSolverTypeOrder(self.msg).solver_type)

    @solver_type.setter
    def solver_type(self, solver_type):
        self.__stub.SetSolverTypeOrder(
            pb.SolverTypePropertyMessage(target=self.msg, solver_type=solver_type.value)
        )

    @property
    def relative_residual(self):
        """:class:`float`: Relative residual value that the HFSS iterative solver is to use."""
        return self.__stub.GetRelativeResidual(self.msg).value

    @relative_residual.setter
    def relative_residual(self, relative_residual):
        self.__stub.SetRelativeResidual(messages.double_property_message(self, relative_residual))

    @property
    def use_shell_elements(self):
        """:class:`bool`: Flag indicating whether to use shell elements."""
        return self.__stub.GetUseShellElements(self.msg).value

    @use_shell_elements.setter
    def use_shell_elements(self, use_shell_elements):
        self.__stub.SetUseShellElements(messages.bool_property_message(self, use_shell_elements))

    @property
    def enhanced_low_frequency_accuracy(self):
        """:obj:`bool`: Flag indicating if enhanced low-frequency accuracy is enabled during simulation."""
        return self.__stub.GetEnhancedLowFrequencyAccuracy(self.msg).value

    @enhanced_low_frequency_accuracy.setter
    def enhanced_low_frequency_accuracy(self, enhanced_low_frequency_accuracy):
        self.__stub.SetEnhancedLowFrequencyAccuracy(
            messages.bool_property_message(self, enhanced_low_frequency_accuracy)
        )


class HFSSSolverSettings(SolverSettings):
    """Representis solver settings for HFSS simulations."""

    __stub: HFSSSolverSettingsServiceStub = StubAccessor(StubType.hfss_solver_sim_settings)

    @property
    def max_delta_z0(self):
        """:obj:`float`: Maximum percent change in characteristic impedance of ports between adaptive passes."""
        return self.__stub.GetMaxDeltaZ0(self.msg).value

    @max_delta_z0.setter
    def max_delta_z0(self, max_delta_z0):
        self.__stub.SetMaxDeltaZ0(messages.double_property_message(self, max_delta_z0))

    @property
    def set_triangles_for_wave_port(self):
        """:obj:`bool`: Flag indicating ifthe minimum and maximum triangle values for waveports are used."""
        return self.__stub.GetSetTrianglesForWaveport(self.msg).value

    @set_triangles_for_wave_port.setter
    def set_triangles_for_wave_port(self, set_triangles_for_wave_port):
        self.__stub.SetSetTrianglesForWaveport(
            messages.bool_property_message(self, set_triangles_for_wave_port)
        )

    @property
    def min_triangles_for_wave_port(self):
        """:obj:`int`: Minimum number of triangles to use for meshing waveports."""
        return self.__stub.GetMinTrianglesForWavePort(self.msg).value

    @min_triangles_for_wave_port.setter
    def min_triangles_for_wave_port(self, min_triangles_for_wave_port):
        self.__stub.SetMinTrianglesForWavePort(
            messages.uint64_property_message(self, min_triangles_for_wave_port)
        )

    @property
    def max_triangles_for_wave_port(self):
        """:obj:`int`: Maximum number of triangles to use for meshing waveports."""
        return self.__stub.GetMaxTrianglesForWavePort(self.msg).value

    @max_triangles_for_wave_port.setter
    def max_triangles_for_wave_port(self, max_triangles_for_wave_port):
        self.__stub.SetMaxTrianglesForWavePort(
            messages.uint64_property_message(self, max_triangles_for_wave_port)
        )

    @property
    def enable_intra_plane_coupling(self):
        """:obj:`bool`: Flag indicating if intra-plane coupling of power/ground nets is enabled to enhance accuracy."""
        return self.__stub.GetIntraPlaneCouplingEnabled(self.msg).value

    @enable_intra_plane_coupling.setter
    def enable_intra_plane_coupling(self, enable_intra_plane_coupling):
        self.__stub.SetIntraPlaneCouplingEnabled(
            messages.bool_property_message(self, enable_intra_plane_coupling)
        )


class HFSSAdvancedSettings(AdvancedSettings):
    """Represents advanced settings for HFSS simulations."""

    __stub: HFSSAdvancedSettingsServiceStub = StubAccessor(StubType.hfss_advanced_sim_settings)

    @property
    def ic_mode_auto_resolution(self):
        """:obj:`bool`: Flag indicating if model resolution is automatically calculated for IC designs."""
        return self.__stub.GetICModeAutoResolution(self.msg).value

    @ic_mode_auto_resolution.setter
    def ic_mode_auto_resolution(self, ic_mode_auto_resolution):
        self.__stub.SetICModeAutoResolution(
            messages.bool_property_message(self, ic_mode_auto_resolution)
        )

    @property
    def ic_mode_length(self):
        """:obj:`str`: Model resolution to use when manually setting the model resolution of IC designs."""
        return self.__stub.GetICModeLength(self.msg).value

    @ic_mode_length.setter
    def ic_mode_length(self, ic_mode_length):
        self.__stub.SetICModeLength(messages.string_property_message(self, ic_mode_length))


class HFSSAdvancedMeshingSettings(AdvancedMeshingSettings):
    """Represents advanced meshing settings for HFSS simulations."""

    __stub: HFSSAdvancedMeshingSettingsServiceStub = StubAccessor(
        StubType.hfss_advanced_sim_meshing_settings
    )

    @property
    def layer_snap_tol(self):
        """:obj:`str`: Snapping tolerance for hierarchical layer alignment.

        Unitless values represent a fraction of the total stackup height.
        """
        return self.__stub.GetLayerAlignment(self.msg).value

    @layer_snap_tol.setter
    def layer_snap_tol(self, layer_snap_tol):
        self.__stub.SetLayerAlignment(messages.string_property_message(self, layer_snap_tol))


class HFSSDCRSettings(SimulationSettingsBase):
    """Represents DCR settings for HFSS simulations."""

    __stub: DCRSettingsServiceStub = StubAccessor(StubType.hfss_dcr_sim_settings)

    @property
    def max_passes(self):
        """:obj:`int`: Maximum number of conduction adaptive passes."""
        return self.__stub.GetMaxPasses(self.msg).value

    @max_passes.setter
    def max_passes(self, max_passes):
        self.__stub.SetMaxPasses(messages.uint64_property_message(self, max_passes))

    @property
    def min_passes(self):
        """:obj:`int`: Minimum number of conduction adaptive passes."""
        return self.__stub.GetMinPasses(self.msg).value

    @min_passes.setter
    def min_passes(self, min_passes):
        self.__stub.SetMinPasses(messages.uint64_property_message(self, min_passes))

    @property
    def min_converged_passes(self):
        """:obj:`int`: Minimum number of converged conduction adaptive passes."""
        return self.__stub.GetMinConvergedPasses(self.msg).value

    @min_converged_passes.setter
    def min_converged_passes(self, min_converged_passes):
        self.__stub.SetMinConvergedPasses(
            messages.uint64_property_message(self, min_converged_passes)
        )

    @property
    def percent_error(self):
        """:obj:`float`: Percent error during conduction adaptive passes."""
        return self.__stub.GetPercentError(self.msg).value

    @percent_error.setter
    def percent_error(self, percent_error):
        self.__stub.SetPercentError(messages.double_property_message(self, percent_error))

    @property
    def percent_refinement_per_pass(self):
        """:obj:`float`: Mesh refinement percentage per conduction adaptive pass."""
        return self.__stub.GetPercentRefinementPerPass(self.msg).value

    @percent_refinement_per_pass.setter
    def percent_refinement_per_pass(self, percent_refinement_per_pass):
        self.__stub.SetPercentRefinementPerPass(
            messages.double_property_message(self, percent_refinement_per_pass)
        )
